"""Tests for OAuth2 Applications."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.forms.models import model_to_dict
from djblets.testing.decorators import add_fixtures

from reviewboard.oauth.forms import (ApplicationChangeForm,
                                     ApplicationCreationForm,
                                     UserApplicationChangeForm,
                                     UserApplicationCreationForm)
from reviewboard.oauth.models import Application
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class ApplicationChangeFormTests(TestCase):
    """Tests for the ApplicationChangeForm."""

    fixtures = ['test_users']

    def test_reassign_client_id(self):
        """Testing ApplicationChangeForm cannot re-assign client_id"""
        user = User.objects.get(username='doc')
        application = self.create_oauth_application(user)
        original_id = application.client_id
        form = ApplicationChangeForm(
            data=dict(
                model_to_dict(
                    instance=application,
                    fields=ApplicationChangeForm.base_fields,
                    exclude=('client_id', 'client_secret')
                ),
                client_id='foo',
            ),
            instance=application,
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.client_id, original_id)

    def test_reassign_client_secret(self):
        """Testing ApplicationChangeForm cannot re-assign client_secret"""
        user = User.objects.get(username='doc')
        application = self.create_oauth_application(user)
        original_secret = application.client_secret
        form = ApplicationChangeForm(
            data=dict(
                model_to_dict(
                    instance=application,
                    fields=ApplicationChangeForm.base_fields,
                    exclude=('client_id', 'client_secret')
                ),
                client_secret='bar',
            ),
            instance=application,
        )
        form.is_valid()
        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.client_secret, original_secret)

    def test_grant_implicit_no_uris(self):
        """Testing ApplicationChangeForm.clean() with GRANT_IMPLICIT and no
        URIs matches AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            '', Application.GRANT_IMPLICIT, False)

    def test_grant_implicit_uris(self):
        """Testing ApplicationChangeForm.clean() with GRANT_IMPLICIT and URIs
        matches AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            'http://example.com/', Application.GRANT_IMPLICIT, True)

    def test_grant_authorization_code_no_uris(self):
        """Testing ApplicationChangeForm.clean() with
        GRANT_AUTHORIZATION_CODE and no URIs matches
        AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            '', Application.GRANT_AUTHORIZATION_CODE, False)

    def test_grant_authorization_code_uris(self):
        """Testing ApplicationChangeForm.clean() with
        GRANT_AUTHORIZATION_CODE and URIS matches AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            'http://example.com/', Application.GRANT_AUTHORIZATION_CODE, True)

    def test_grant_password_no_uris(self):
        """Testing ApplicationChangeForm.clean() with GRANT_PASSWORD and no
        URIs matches AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            '', Application.GRANT_PASSWORD, True)

    def test_grant_password_uris(self):
        """Testing ApplicationChangeForm.clean() with GRANT_PASSWORD and URIs
        matches AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            'http://example.com/', Application.GRANT_PASSWORD, True)

    def test_grant_client_credentials_no_uris(self):
        """Testing ApplicationChangeForm.clean() with
        GRANT_CLIENT_CREDENTIALS and no URIs matches
        AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            '', Application.GRANT_CLIENT_CREDENTIALS, True)

    def test_grant_client_credentials_uris(self):
        """Testing ApplicationChangeForm.clean() with
        GRANT_CLIENT_CREDENTIALS and no URIs matches
        AbstractApplication.clean()
        """
        self._test_redirect_uri_grant_combination(
            '', Application.GRANT_CLIENT_CREDENTIALS, True)

    @add_fixtures(['test_site'])
    def test_enable_disabled_for_security(self):
        """Testing ApplicationChangeForm will not enable an application
        disabled for security
        """
        local_site = LocalSite.objects.get(pk=1)
        admin = User.objects.get(username='admin')
        owner = User.objects.get(username='doc')
        local_site.users.remove(owner)

        application = self.create_oauth_application(user=admin,
                                                    local_site=local_site,
                                                    enabled=False,
                                                    original_user=owner)

        self.assertTrue(application.is_disabled_for_security)
        self.assertEqual(application.original_user, owner)

        form = ApplicationChangeForm(
            data=dict(model_to_dict(application),
                      enabled=True),
            instance=application,
        )

        self.assertFalse(form.is_valid())
        self.assertEqual(form.non_field_errors(),
                         [ApplicationCreationForm.DISABLED_FOR_SECURITY_ERROR])

    def _test_redirect_uri_grant_combination(self, redirect_uris, grant_type,
                                             is_valid):
        doc = User.objects.get(username='doc')
        common_fields = {
            'authorization_grant_type': grant_type,
            'redirect_uris': redirect_uris,
        }

        application = self.create_oauth_application(user=doc)

        # This should always succeed.
        super(Application, application).clean()

        form = ApplicationChangeForm(
            data=dict(model_to_dict(application), **common_fields),
            instance=application,
        )

        self.assertEqual(form.is_valid(), is_valid)

        application = Application(user=doc, **common_fields)

        # Ensure that the error cases of AbstractApplication.clean() matches
        # our implementation.
        if is_valid:
            super(Application, application).clean()
        else:
            self.assertIn('redirect_uris', form.errors)

            with self.assertRaises(ValidationError):
                super(Application, application).clean()


class ApplicationCreationFormTests(TestCase):
    """Tests for the ApplicationCreationForm."""

    fixtures = ['test_users']

    def test_valid_client_id_and_secret(self):
        """Testing ApplicationCreationForm sets a valid client_id and
        client_secret
        """
        form = ApplicationCreationForm(data={
            'authorization_grant_type': Application.GRANT_CLIENT_CREDENTIALS,
            'client_id': 'foo',
            'client_type': Application.CLIENT_PUBLIC,
            'client_secret': 'bar',
            'enabled': True,
            'name': 'Test Application',
            'redirect_uris': '',
            'user': 1,
        })

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertNotEqual(application.client_id, form.data['client_id'])
        self.assertNotEqual(application.client_secret,
                            form.data['client_secret'])
        self.assertGreater(len(application.client_id), 0)
        self.assertGreater(len(application.client_secret), 0)


class UserApplicationCreationFormTests(TestCase):
    """Tests for the UserApplicationCreationForm."""

    fixtures = ['test_users']

    def test_set_user(self):
        """Testing UserApplicationCreationForm cannot assign different user"""
        user = User.objects.get(username='doc')
        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
                'user': 2,
            },
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.user, user)

    @add_fixtures(['test_site'])
    def test_assign_local_site(self):
        """Testing UserApplicationCreationForm with Local Site"""
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.get(name=self.local_site_name)

        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
                'local_site': local_site.pk
            },
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.local_site, local_site)

    def test_assign_local_site_inacessible(self):
        """Testing UserApplicationCreationForm with an inaccessible Local Site
        """
        local_site = LocalSite.objects.create(name='inacessible')
        user = User.objects.get(username='doc')

        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
                'local_site': local_site.pk
            },
        )

        self.assertFalse(form.is_valid())

    def test_set_extra_data(self):
        """Testing UserApplicationCreationForm cannot assign extra_data"""
        user = User.objects.get(username='doc')
        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
                'extra_data': 1,
            },
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.extra_data, {})

    def test_set_skip_authorization(self):
        """Testing UserApplicationCreationForm cannot assign
        skip_authorization
        """
        user = User.objects.get(username='doc')
        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
                'extra_data': 1,
            },
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.skip_authorization, False)

    def test_set_client_id(self):
        """Testing UserApplicationCreationForm cannot assign client_id
        """
        user = User.objects.get(username='doc')
        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_id': 'foo',
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
            },
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertNotEqual(application.client_id, 'foo')
        self.assertNotEqual(len(application.client_id), 0)

    def test_set_client_secret(self):
        """Testing UserApplicationCreationForm cannot assign client_secret
        """
        user = User.objects.get(username='doc')
        form = UserApplicationCreationForm(
            user,
            data={
                'authorization_grant_type': Application.GRANT_IMPLICIT,
                'client_secret': 'bar',
                'client_type': Application.CLIENT_PUBLIC,
                'name': 'test',
                'redirect_uris': 'http://example.com',
            },
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertNotEqual(application.client_secret, 'bar')
        self.assertNotEqual(len(application.client_secret), 0)


class UserApplicationChangeFormTests(TestCase):
    """Tests for the UserApplicationChangeForm."""

    fixtures = ['test_users']

    def test_reassign_user(self):
        """Testing UserApplicationChangeForm cannot re-assign different user"""
        user = User.objects.get(username='doc')
        application = self.create_oauth_application(user)
        form = UserApplicationChangeForm(
            user,
            data=dict(
                model_to_dict(
                    instance=application,
                    fields=UserApplicationChangeForm.base_fields,
                    exclude=('client_id', 'client_secret'),
                ),
                user=2,
            ),
            instance=application,
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.user, user)

    @add_fixtures(['test_site'])
    def test_reassign_local_site(self):
        """Testing UserApplicationChangeForm cannot re-assign Local Site"""
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.get(pk=1)
        application = self.create_oauth_application(user, local_site)

        form = UserApplicationChangeForm(
            user,
            data=dict(
                model_to_dict(
                    instance=application,
                    fields=UserApplicationChangeForm.base_fields,
                    exclude=('client_id', 'client_secret'),
                ),
                local_site=2,
            ),
            instance=application,
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.local_site, local_site)

    def test_reassign_extra_data(self):
        """Testing UserApplicationChangeForm cannot re-assign extra_data"""
        user = User.objects.get(username='doc')
        application = self.create_oauth_application(user)
        form = UserApplicationChangeForm(
            user,
            data=dict(
                model_to_dict(
                    instance=application,
                    fields=UserApplicationChangeForm.base_fields,
                    exclude=('client_id', 'client_secret'),
                ),
                extra_data=1,
            ),
            instance=application,
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.extra_data, {})

    def test_reassign_skip_authorization(self):
        """Testing UserApplicationChangeForm cannot re-assign
        skip_authorization
        """
        user = User.objects.get(username='doc')
        application = self.create_oauth_application(user)
        form = UserApplicationChangeForm(
            user,
            data=dict(
                model_to_dict(
                    instance=application,
                    fields=UserApplicationChangeForm.base_fields,
                    exclude=('client_id', 'client_secret'),
                ),
                skip_authorization=True,
            ),
            instance=application,
        )

        self.assertTrue(form.is_valid())
        application = form.save()
        self.assertEqual(application.skip_authorization, False)



"""Account-related template tags."""

from __future__ import unicode_literals

from django import template
from django.utils.html import escape


register = template.Library()


@register.simple_tag(takes_context=True)
def user_profile_display_name(context, user):
    """Render the user's display name.

    Args:
        context (django.template.context.Context):
            The template rendering context.

        user (django.contrib.auth.models.User):
            The user whose display name is to be rendered.

    Returns:
        unicode:
        The user's display name.
    """
    request = context['request']

    return escape(user.get_profile().get_display_name(request.user))



from __future__ import unicode_literals

import re
import sre_constants

from django import forms
from django.contrib.auth.models import User
from django.contrib.auth.forms import \
    AuthenticationForm as DjangoAuthenticationForm
from django.core.exceptions import ValidationError
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _
from djblets.auth.ratelimit import is_ratelimited
from djblets.siteconfig.forms import SiteSettingsForm

from reviewboard.admin.checks import get_can_enable_dns, get_can_enable_ldap


class ActiveDirectorySettingsForm(SiteSettingsForm):
    """A form for configuring the Active Directory authentication backend."""

    auth_ad_domain_name = forms.CharField(
        label=_("Domain name"),
        help_text=_("Enter the domain name to use, (ie. example.com). This "
                    "will be used to query for LDAP servers and to bind to "
                    "the domain."),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ad_use_tls = forms.BooleanField(
        label=_("Use TLS for authentication"),
        required=False)

    auth_ad_find_dc_from_dns = forms.BooleanField(
        label=_("Find DC from DNS"),
        help_text=_("Query DNS to find which domain controller to use"),
        required=False)

    auth_ad_domain_controller = forms.CharField(
        label=_("Domain controller"),
        help_text=_("If not using DNS to find the DC, specify the domain "
                    "controller(s) here "
                    "(eg. ctrl1.example.com ctrl2.example.com:389)"),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ad_ou_name = forms.CharField(
        label=_("OU name"),
        help_text=_("Optionally restrict users to specified OU."),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ad_group_name = forms.CharField(
        label=_("Group name"),
        help_text=_("Optionally restrict users to specified group."),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ad_search_root = forms.CharField(
        label=_("Custom search root"),
        help_text=_("Optionally specify a custom search root, overriding "
                    "the built-in computed search root. If set, \"OU name\" "
                    "is ignored."),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ad_recursion_depth = forms.IntegerField(
        label=_("Recursion Depth"),
        help_text=_('Depth to recurse when checking group membership. '
                    '0 to turn off, -1 for unlimited.'),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    def load(self):
        """Load the data for the form."""
        can_enable_dns, reason = get_can_enable_dns()

        if not can_enable_dns:
            self.disabled_fields['auth_ad_find_dc_from_dns'] = reason

        can_enable_ldap, reason = get_can_enable_ldap()

        if not can_enable_ldap:
            self.disabled_fields['auth_ad_use_tls'] = True
            self.disabled_fields['auth_ad_group_name'] = True
            self.disabled_fields['auth_ad_recursion_depth'] = True
            self.disabled_fields['auth_ad_ou_name'] = True
            self.disabled_fields['auth_ad_search_root'] = True
            self.disabled_fields['auth_ad_find_dc_from_dns'] = True
            self.disabled_fields['auth_ad_domain_controller'] = True

            self.disabled_reasons['auth_ad_domain_name'] = reason

        super(ActiveDirectorySettingsForm, self).load()

    class Meta:
        title = _('Active Directory Authentication Settings')
        fieldsets = (
            (None, {
                'fields': ('auth_ad_domain_name',
                           'auth_ad_use_tls',
                           'auth_ad_find_dc_from_dns',
                           'auth_ad_domain_controller'),
            }),
            (_('Access Control Settings'), {
                'fields': ('auth_ad_ou_name',
                           'auth_ad_group_name'),
            }),
            (_('Advanced Settings'), {
                'fields': ('auth_ad_search_root',
                           'auth_ad_recursion_depth'),
            }),
        )


class StandardAuthSettingsForm(SiteSettingsForm):
    """A form for configuring the builtin authentication backend."""

    auth_enable_registration = forms.BooleanField(
        label=_("Enable registration"),
        help_text=_("Allow users to register new accounts."),
        required=False)

    auth_registration_show_captcha = forms.BooleanField(
        label=_('Show a captcha for registration'),
        required=False)

    recaptcha_public_key = forms.CharField(
        label=_('reCAPTCHA Public Key'),
        required=False,
        widget=forms.TextInput(attrs={'size': '60'}))

    recaptcha_private_key = forms.CharField(
        label=_('reCAPTCHA Private Key'),
        required=False,
        widget=forms.TextInput(attrs={'size': '60'}))

    def __init__(self, *args, **kwargs):
        super(StandardAuthSettingsForm, self).__init__(*args, **kwargs)

        # This is done at initialization time instead of in the field
        # definition because the format operation causes the lazy translation
        # to be evaluated, which has to happen after the i18n infrastructure
        # has been started.
        self.fields['auth_registration_show_captcha'].help_text = mark_safe(
            _('Displays a captcha using <a href="%(recaptcha_url)s">'
              'reCAPTCHA</a> on the registration page. To enable this, you '
              'will need to go <a href="%(register_url)s">here</A> to '
              'register an account and type in your new keys below.')
            % {
                'recaptcha_url': 'http://www.google.com/recaptcha',
                'register_url': 'https://www.google.com/recaptcha/admin'
                                '#createsite',
            })

    def clean_recaptcha_public_key(self):
        """Validate that the reCAPTCHA public key is specified if needed."""
        key = self.cleaned_data['recaptcha_public_key'].strip()

        if self.cleaned_data['auth_registration_show_captcha'] and not key:
            raise ValidationError(_('This field is required.'))

        return key

    def clean_recaptcha_private_key(self):
        """Validate that the reCAPTCHA private key is specified if needed."""
        key = self.cleaned_data['recaptcha_private_key'].strip()

        if self.cleaned_data['auth_registration_show_captcha'] and not key:
            raise ValidationError(_('This field is required.'))

        return key

    class Meta:
        title = _('Registration Settings')
        fieldsets = (
            (None, {
                'fields': ('auth_enable_registration',),
            }),
            (_('reCAPTCHA Settings'), {
                'fields': ('auth_registration_show_captcha',
                           'recaptcha_public_key',
                           'recaptcha_private_key'),
            }),
        )


class HTTPBasicSettingsForm(SiteSettingsForm):
    """A form for configuring the HTTP Digest authentication backend."""

    auth_digest_file_location = forms.CharField(
        label=_(".htpasswd File location"),
        help_text=_("Location of the .htpasswd file which "
                    "stores the usernames and passwords in digest format"),
        widget=forms.TextInput(attrs={'size': '60'}))

    auth_digest_realm = forms.CharField(
        label=_("HTTP Digest Realm"),
        help_text=_("Realm used for HTTP Digest authentication"),
        widget=forms.TextInput(attrs={'size': '40'}))

    class Meta:
        title = _('HTTP Digest Authentication Settings')


class LDAPSettingsForm(SiteSettingsForm):
    """A form for configuring the LDAP authentication backend."""

    # TODO: Invent a URIField and use it.
    auth_ldap_uri = forms.CharField(
        label=_("LDAP Server"),
        help_text=_("The LDAP server to authenticate with. "
                    "For example: ldap://localhost:389"),
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ldap_base_dn = forms.CharField(
        label=_("LDAP Base DN"),
        help_text=_("The LDAP Base DN for performing LDAP searches.  For "
                    "example: ou=users,dc=example,dc=com"),
        required=True,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ldap_uid = forms.CharField(
        label=_("Username Attribute"),
        help_text=_("The attribute in the LDAP server that stores a user's "
                    "login name."),
        required=True)

    auth_ldap_given_name_attribute = forms.CharField(
        label=_("Given Name Attribute"),
        initial="givenName",
        help_text=_("The attribute in the LDAP server that stores the user's "
                    "given name."),
        required=False)

    auth_ldap_surname_attribute = forms.CharField(
        label=_("Surname Attribute"),
        initial="sn",
        help_text=_("The attribute in the LDAP server that stores the user's "
                    "surname."),
        required=False)

    auth_ldap_full_name_attribute = forms.CharField(
        label=_("Full Name Attribute"),
        help_text=_("The attribute in the LDAP server that stores the user's "
                    "full name.  This takes precedence over the "
                    '"Given Name Attribute" and "Surname Attribute."'),
        required=False)

    auth_ldap_email_domain = forms.CharField(
        label=_("E-Mail Domain"),
        help_text=_("The domain name appended to the username to construct "
                    "the user's e-mail address. This takes precedence over "
                    '"E-Mail LDAP Attribute."'),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ldap_email_attribute = forms.CharField(
        label=_("E-Mail LDAP Attribute"),
        help_text=_("The attribute in the LDAP server that stores the user's "
                    "e-mail address. For example: mail"),
        required=False)

    auth_ldap_tls = forms.BooleanField(
        label=_("Use TLS for authentication"),
        required=False)

    auth_ldap_uid_mask = forms.CharField(
        label=_("Custom LDAP User Search Filter"),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ldap_anon_bind_uid = forms.CharField(
        label=_("Review Board LDAP Bind Account"),
        help_text=_("The full distinguished name of a user account with "
                    "sufficient access to perform lookups of users and "
                    "groups in the LDAP server. If the LDAP server permits "
                    "such lookups via anonymous bind, you may leave this "
                    "field blank."),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_ldap_anon_bind_passwd = forms.CharField(
        label=_("Review Board LDAP Bind Password"),
        widget=forms.PasswordInput(attrs={'size': '30'}, render_value=True),
        help_text=_("The password for the Review Board LDAP Bind Account."),
        required=False)

    def __init__(self, *args, **kwargs):
        super(LDAPSettingsForm, self).__init__(*args, **kwargs)

        # This is done at initialization time instead of in the field
        # definition because the format operation causes the lazy translation
        # to be evaluated, which has to happen after the i18n infrastructure
        # has been started.
        self.fields['auth_ldap_uid_mask'].help_text = mark_safe(
            _('A custom LDAP search filter, corresponding to RFC 2254. If '
              'left unset, this option is equivalent to <tt>('
              'usernameattribute=%(varname)s)</tt>. Use <tt>'
              '"%(varname)s"</tt> wherever the username would normally go. '
              'Specify this value only if the default cannot locate all '
              'users.')
            % {
                'varname': '%s',
            })

    def load(self):
        """Load the data for the form."""
        can_enable_ldap, reason = get_can_enable_ldap()

        if not can_enable_ldap:
            self.disabled_fields['auth_ldap_uri'] = True
            self.disabled_fields['auth_ldap_given_name_attribute'] = True
            self.disabled_fields['auth_ldap_surname_attribute'] = True
            self.disabled_fields['auth_ldap_full_name_attribute'] = True
            self.disabled_fields['auth_ldap_email_domain'] = True
            self.disabled_fields['auth_ldap_email_attribute'] = True
            self.disabled_fields['auth_ldap_tls'] = True
            self.disabled_fields['auth_ldap_base_dn'] = True
            self.disabled_fields['auth_ldap_uid'] = True
            self.disabled_fields['auth_ldap_uid_mask'] = True
            self.disabled_fields['auth_ldap_anon_bind_uid'] = True
            self.disabled_fields['auth_ldap_anon_bind_password'] = True

            self.disabled_reasons['auth_ldap_uri'] = reason

        super(LDAPSettingsForm, self).load()

    class Meta:
        title = _('LDAP Authentication Settings')
        fieldsets = (
            (None, {
                'fields': ('auth_ldap_uri',
                           'auth_ldap_tls',
                           'auth_ldap_anon_bind_uid',
                           'auth_ldap_anon_bind_passwd',
                           'auth_ldap_base_dn'),
            }),
            (_('User Lookups'), {
                'fields': ('auth_ldap_uid',
                           'auth_ldap_given_name_attribute',
                           'auth_ldap_surname_attribute',
                           'auth_ldap_full_name_attribute',
                           'auth_ldap_email_attribute',
                           'auth_ldap_email_domain',
                           'auth_ldap_uid_mask'),
            }),
        )


class LegacyAuthModuleSettingsForm(SiteSettingsForm):
    """A form for configuring old-style custom authentication backends.

    Newer authentication backends are registered via the extensions framework,
    but there used to be a method by which users just put in a list of python
    module paths. This form allows that configuration to be edited.
    """

    custom_backends = forms.CharField(
        label=_("Backends"),
        help_text=_('A comma-separated list of old-style custom auth '
                    'backends. These are represented as Python module paths.'),
        widget=forms.TextInput(attrs={'size': '40'}))

    def load(self):
        """Load the data for the form."""
        self.fields['custom_backends'].initial = \
            ', '.join(self.siteconfig.get('auth_custom_backends'))

        super(LegacyAuthModuleSettingsForm, self).load()

    def save(self):
        """Save the form."""
        self.siteconfig.set(
            'auth_custom_backends',
            re.split(r',\s*', self.cleaned_data['custom_backends']))

        super(LegacyAuthModuleSettingsForm, self).save()

    class Meta:
        title = _('Legacy Authentication Module Settings')
        save_blacklist = ('custom_backends',)


class NISSettingsForm(SiteSettingsForm):
    """A form for configuring the NIS authentication backend."""

    auth_nis_email_domain = forms.CharField(
        label=_("E-Mail Domain"),
        widget=forms.TextInput(attrs={'size': '40'}))

    class Meta:
        title = _('NIS Authentication Settings')


class X509SettingsForm(SiteSettingsForm):
    """A form for configuring the X509 certificate authentication backend."""

    auth_x509_username_field = forms.ChoiceField(
        label=_("Username Field"),
        choices=(
            # Note: These names correspond to environment variables set by
            #       mod_ssl.
            ("SSL_CLIENT_S_DN", _("DN (Distinguished Name)")),
            ("SSL_CLIENT_S_DN_CN", _("CN (Common Name)")),
            ("SSL_CLIENT_S_DN_Email", _("E-mail address")),
        ),
        help_text=_("The X.509 certificate field from which the Review Board "
                    "username will be extracted."),
        required=True)

    auth_x509_username_regex = forms.CharField(
        label=_("Username Regex"),
        help_text=_("Optional regex used to convert the selected X.509 "
                    "certificate field to a usable Review Board username. For "
                    "example, if using the e-mail field to retrieve the "
                    "username, use this regex to get the username from an "
                    "e-mail address: '(\s+)@yoursite.com'. There must be only "
                    "one group in the regex."),
        required=False,
        widget=forms.TextInput(attrs={'size': '40'}))

    auth_x509_autocreate_users = forms.BooleanField(
        label=_("Automatically create new user accounts."),
        help_text=_("Enabling this option will cause new user accounts to be "
                    "automatically created when a new user with an X.509 "
                    "certificate accesses Review Board."),
        required=False)

    def clean_auth_x509_username_regex(self):
        """Validate that the specified regular expression is valid."""
        regex = self.cleaned_data['auth_x509_username_regex']

        try:
            re.compile(regex)
        except sre_constants.error as e:
            raise ValidationError(e)

        return regex

    class Meta:
        title = _('X.509 Client Certificate Authentication Settings')


class AuthenticationForm(DjangoAuthenticationForm):
    """Form used for user logins.

    This extends Django's built-in AuthenticationForm implementation to allow
    users to specify their e-mail address in place of their username. In
    addition, it also tracks the number of failed login attempts for a given
    time frame, and informs the user whether the maximum number of attempts
    have been exceeded.
    """

    username = forms.CharField(
        label=_("Username"),
        widget=forms.TextInput(attrs={'autofocus': 'autofocus'}))

    def clean_username(self):
        """Validate the 'username' field.

        In case the given text is not a user found on the system, attempt a
        look-up using it as an e-mail address and change the user-entered text
        so that login can succeed.
        """
        username = self.cleaned_data.get('username')

        if not User.objects.filter(username=username).exists():
            try:
                username = User.objects.get(email=username).username
            except (User.DoesNotExist, User.MultipleObjectsReturned):
                pass

        return username

    def clean(self):
        """Validate the authentication form.

        In case authentication has failed for the given user, Djblets's rate
        limiting feature will increment the number of failed login attempts
        until the maximum number of attempts have been reached. The user
        will have to wait until the rate limit time period is over before
        trying again.

        Returns:
            dict:
            The cleaned data for all fields in the form.

        Raises:
            django.core.exceptions.ValidationError:
                The data in the form was not valid.
        """
        request = self.request

        # Check if the number of failed login attempts were already exceeded
        # before authenticating.
        if is_ratelimited(request, increment=False):
            raise forms.ValidationError(
                _('Maximum number of login attempts exceeded.'))

        try:
            self.cleaned_data = super(AuthenticationForm, self).clean()
        except ValidationError:
            # If authentication for a given user has failed (i.e.
            # self.user_cache is None), increment the number of
            # failed login attempts.
            is_ratelimited(request, increment=True)
            raise

        return self.cleaned_data



from __future__ import unicode_literals

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from djblets.db.fields import JSONField
from djblets.util.compat.django.core.validators import URLValidator
from multiselectfield import MultiSelectField

from reviewboard.notifications.managers import WebHookTargetManager
from reviewboard.scmtools.models import Repository
from reviewboard.site.models import LocalSite


@python_2_unicode_compatible
class WebHookTarget(models.Model):
    """A target for a webhook.

    A webhook target is a URL which will receive a POST request when the
    corresponding event occurs.
    """
    ALL_EVENTS = '*'

    EVENT_CHOICES = (
        (ALL_EVENTS, _('All events')),
        ('review_request_closed', _('Review request closed')),
        ('review_request_published', _('Review request published')),
        ('review_request_reopened', _('Review request reopened')),
        ('review_published', _('Review published')),
        ('reply_published', _('Reply published')),
    )

    ENCODING_JSON = 'application/json'
    ENCODING_XML = 'application/xml'
    ENCODING_FORM_DATA = 'application/x-www-form-urlencoded'

    ALL_ENCODINGS = (
        ENCODING_JSON,
        ENCODING_XML,
        ENCODING_FORM_DATA
    )

    ENCODINGS = (
        (ENCODING_JSON, _('JSON')),
        (ENCODING_XML, _('XML')),
        (ENCODING_FORM_DATA, _('Form Data')),
    )

    APPLY_TO_ALL = 'A'
    APPLY_TO_NO_REPOS = 'N'
    APPLY_TO_SELECTED_REPOS = 'S'

    APPLY_TO_CHOICES = (
        (APPLY_TO_ALL, _('All review requests')),
        (APPLY_TO_SELECTED_REPOS,
         _('Only review requests on selected repositories')),
        (APPLY_TO_NO_REPOS,
         _('Only review requests not associated with a repository (file '
           'attachments only)')),
    )

    # Standard information
    enabled = models.BooleanField(default=True)
    events = MultiSelectField(
        _('events'),
        choices=EVENT_CHOICES,
        blank=True,
        help_text=_('Select any or all events that should trigger this '
                    'Webhook.'))

    url = models.URLField(
        'URL',
        help_text=_('When the event is triggered, HTTP requests will be '
                    'made against this URL.'))

    encoding = models.CharField(
        _('encoding'),
        choices=ENCODINGS,
        default=ENCODING_JSON,
        max_length=40,
        help_text=_('Payload contents will be encoded in this format.'))

    # Custom content
    use_custom_content = models.BooleanField(
        _('use custom payload content'),
        default=False)

    custom_content = models.TextField(
        _('custom content'),
        blank=True,
        null=True,
        help_text=_('You can override what is sent to the URL above. If '
                    'left blank, the default payload will be sent.'))

    # HMAC payload signing
    secret = models.CharField(
        _('HMAC secret'),
        max_length=128, blank=True,
        help_text=_('If specified, the HMAC digest for the Webhook payload '
                    'will be signed with the given secret.'))

    # Apply to
    apply_to = models.CharField(
        _('apply to'),
        max_length=1,
        blank=False,
        default=APPLY_TO_ALL,
        choices=APPLY_TO_CHOICES)

    repositories = models.ManyToManyField(
        Repository,
        blank=True,
        null=True,
        related_name='webhooks',
        help_text=_('If set, this Webhook will be limited to these '
                    'repositories.'))

    local_site = models.ForeignKey(
        LocalSite,
        blank=True,
        null=True,
        related_name='webhooks',
        help_text=_('If set, this Webhook will be limited to this site.'))

    extra_data = JSONField(
        null=True,
        help_text=_('Extra JSON data that can be tied to this Webhook '
                    'registration. It will not be sent with the Webhook '
                    'request.'))

    objects = WebHookTargetManager()

    def __init__(self, *args, **kwargs):
        """Initialize the model.

        Args:
            *args (tuple):
                Positional arguments to pass through to the Model constructor.

            **kwargs (dict):
                Keyword arguments to pass through to the Model constructor.
        """
        super(WebHookTarget, self).__init__(*args, **kwargs)
        self._meta.get_field('url').validators = [URLValidator()]

    def is_accessible_by(self, user, local_site=None):
        """Return if the webhook can be accessed or modified by the user.

        All superusers and admins of the webhook's local site can access and
        modify the webhook.

        Args:
            user (django.contrib.auth.models.User):
                The user who is trying to access the webhook.

            local_site (reviewboard.site.models.LocalSite):
                The current local site, if it exists.

        Returns:
            bool:
            Whether or not the given user can access or modify the webhook
            through the given local site.
        """
        return (user.is_superuser or
                (user.is_authenticated() and
                 local_site and
                 self.local_site_id == local_site.pk and
                 local_site.is_mutable_by(user)))

    def __str__(self):
        return self.url

    class Meta:
        db_table = 'notifications_webhooktarget'
        verbose_name = _('Webhook')
        verbose_name_plural = _('Webhooks')



"""Unit tests for the Bitbucket hosting service."""

from __future__ import unicode_literals

from djblets.testing.decorators import add_fixtures

from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            RepositoryError)
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError
from reviewboard.site.models import LocalSite
from reviewboard.site.urlresolvers import local_site_reverse


class BitbucketTestCase(HostingServiceTestCase):
    """Base class for Bitbucket test suites."""

    service_name = 'bitbucket'
    fixtures = ['test_scmtools']

    default_account_data = {
        'password': encrypt_password(HostingServiceTestCase.default_password),
    }

    default_repository_extra_data = {
        'bitbucket_repo_name': 'myrepo',
    }


class BitbucketTests(BitbucketTestCase):
    """Unit tests for the Bitbucket hosting service."""

    def test_service_support(self):
        """Testing Bitbucket service support capabilities"""
        self.assertTrue(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)

    def test_get_repository_fields_with_git_and_personal_plan(self):
        """Testing Bitbucket.get_repository_fields for Git and plan=personal"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'bitbucket_repo_name': 'myrepo',
                },
                plan='personal'
            ),
            {
                'path': 'git@bitbucket.org:myuser/myrepo.git',
                'mirror_path': ('https://myuser@bitbucket.org/myuser/'
                                'myrepo.git'),
            })

    def test_get_repository_fields_with_mercurial_and_personal_plan(self):
        """Testing Bitbucket.get_repository_fields for Mercurial and
        plan=personal
        """
        self.assertEqual(
            self.get_repository_fields(
                'Mercurial',
                fields={
                    'bitbucket_repo_name': 'myrepo',
                },
                plan='personal'
            ),
            {
                'path': 'https://myuser@bitbucket.org/myuser/myrepo',
                'mirror_path': 'ssh://hg@bitbucket.org/myuser/myrepo',
            })

    def test_get_repository_fields_with_git_and_team_plan(self):
        """Testing Bitbucket.get_repository_fields for Git and plan=team"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'bitbucket_team_name': 'myteam',
                    'bitbucket_team_repo_name': 'myrepo',
                },
                plan='team'
            ),
            {
                'path': 'git@bitbucket.org:myteam/myrepo.git',
                'mirror_path': ('https://myuser@bitbucket.org/myteam/'
                                'myrepo.git'),
            })

    def test_get_repository_fields_with_mercurial_and_team_plan(self):
        """Testing Bitbucket.get_repository_fields for Mercurial and plan=team
        """
        self.assertEqual(
            self.get_repository_fields(
                'Mercurial',
                fields={
                    'bitbucket_team_name': 'myteam',
                    'bitbucket_team_repo_name': 'myrepo',
                },
                plan='team'
            ),
            {
                'path': 'https://myuser@bitbucket.org/myteam/myrepo',
                'mirror_path': 'ssh://hg@bitbucket.org/myteam/myrepo',
            })

    def test_get_repository_fields_with_git_and_other_user_plan(self):
        """Testing Bitbucket.get_repository_fields for Git and plan=other-user
        """
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'bitbucket_other_user_username': 'someuser',
                    'bitbucket_other_user_repo_name': 'myrepo',
                },
                plan='other-user'
            ),
            {
                'path': 'git@bitbucket.org:someuser/myrepo.git',
                'mirror_path': ('https://myuser@bitbucket.org/someuser/'
                                'myrepo.git'),
            })

    def test_get_repository_fields_with_mercurial_and_other_user_plan(self):
        """Testing Bitbucket.get_repository_fields for Mercurial and
        plan=other-user
        """
        self.assertEqual(
            self.get_repository_fields(
                'Mercurial',
                fields={
                    'bitbucket_other_user_username': 'someuser',
                    'bitbucket_other_user_repo_name': 'myrepo',
                },
                plan='other-user'
            ),
            {
                'path': 'https://myuser@bitbucket.org/someuser/myrepo',
                'mirror_path': 'ssh://hg@bitbucket.org/someuser/myrepo',
            })

    def test_get_bug_tracker_field_with_personal_plan(self):
        """Testing Bitbucket.get_bug_tracker_field with plan=personal"""
        self.assertTrue(self.service_class.get_bug_tracker_requires_username(
            plan='personal'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field(
                'personal',
                {
                    'bitbucket_repo_name': 'myrepo',
                    'hosting_account_username': 'myuser',
                }),
            'https://bitbucket.org/myuser/myrepo/issue/%s/')

    def test_get_bug_tracker_field_with_team_plan(self):
        """Testing Bitbucket.get_bug_tracker_field with plan=team"""
        self.assertFalse(self.service_class.get_bug_tracker_requires_username(
            plan='team'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field(
                'team',
                {
                    'bitbucket_team_name': 'myteam',
                    'bitbucket_team_repo_name': 'myrepo',
                }),
            'https://bitbucket.org/myteam/myrepo/issue/%s/')

    def test_get_bug_tracker_field_with_other_user_plan(self):
        """Testing Bitbucket.get_bug_tracker_field with plan=other-user"""
        self.assertFalse(self.service_class.get_bug_tracker_requires_username(
            plan='other-user'))
        self.assertEqual(
            self.service_class.get_bug_tracker_field(
                'other-user',
                {
                    'bitbucket_other_user_username': 'someuser',
                    'bitbucket_other_user_repo_name': 'myrepo',
                }),
            'https://bitbucket.org/someuser/myrepo/issue/%s/')

    def test_check_repository_with_personal_plan(self):
        """Testing Bitbucket.check_repository with plan=personal"""
        with self.setup_http_test(payload=b'{"scm": "git"}',
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(bitbucket_repo_name='myrepo',
                                         plan='personal',
                                         tool_name='Git')

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo'
                 '?fields=scm'))

    def test_check_repository_with_team_plan(self):
        """Testing Bitbucket.check_repository with plan=team"""
        with self.setup_http_test(payload=b'{"scm": "git"}',
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(bitbucket_team_name='myteam',
                                         bitbucket_team_repo_name='myrepo',
                                         tool_name='Git',
                                         plan='team')

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myteam/myrepo'
                 '?fields=scm'))

    def test_check_repository_with_other_user_plan(self):
        """Testing Bitbucket.check_repository with plan=other-user"""
        with self.setup_http_test(payload=b'{"scm": "git"}',
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(
                bitbucket_other_user_username='someuser',
                bitbucket_other_user_repo_name='myrepo',
                plan='other-user',
                tool_name='Git')

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/someuser/myrepo'
                 '?fields=scm'))

    def test_check_repository_with_slash(self):
        """Testing Bitbucket.check_repository with /"""
        expected_message = \
            'Please specify just the name of the repository, not a path.'

        with self.setup_http_test(expected_http_calls=0) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(
                    bitbucket_team_name='myteam',
                    bitbucket_team_repo_name='myteam/myrepo',
                    plan='team')

    def test_check_repository_with_dot_git(self):
        """Testing Bitbucket.check_repository with .git"""
        expected_message = \
            'Please specify just the name of the repository without ".git".'

        with self.setup_http_test(expected_http_calls=0) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(
                    bitbucket_team_name='myteam',
                    bitbucket_team_repo_name='myrepo.git',
                    plan='team')

    def test_check_repository_with_type_mismatch(self):
        """Testing Bitbucket.check_repository with type mismatch"""
        error_message = (
            'The Bitbucket repository being configured does not match the '
            'type of repository you have selected.'
        )

        with self.setup_http_test(payload=b'{"scm": "git"}',
                                  expected_http_calls=1) as ctx:
            # Check Git repositories.
            with self.assertRaisesMessage(RepositoryError, error_message):
                ctx.service.check_repository(
                    bitbucket_team_name='myteam',
                    bitbucket_team_repo_name='myrepo',
                    plan='team',
                    tool_name='Mercurial')

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myteam/myrepo'
                 '?fields=scm'))

        # Now check Mercurial repositories.
        with self.setup_http_test(payload=b'{"scm": "hg"}',
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(RepositoryError, error_message):
                ctx.service.check_repository(
                    bitbucket_team_name='myteam',
                    bitbucket_team_repo_name='myrepo',
                    plan='team',
                    tool_name='Git')

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myteam/myrepo'
                 '?fields=scm'))

    def test_authorize(self):
        """Testing Bitbucket.authorize"""
        hosting_account = self.create_hosting_account(data={})

        with self.setup_http_test(payload=b'{}',
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.is_authorized())
            ctx.service.authorize(username='myuser',
                                  password='abc123')

        self.assertIn('password', hosting_account.data)
        self.assertNotEqual(hosting_account.data['password'], 'abc123')
        self.assertEqual(decrypt_password(hosting_account.data['password']),
                         'abc123')
        self.assertTrue(ctx.service.is_authorized())

        ctx.assertHTTPCall(
            0,
            url='https://bitbucket.org/api/2.0/user',
            username='myuser',
            password='abc123')

    def test_authorize_with_bad_credentials(self):
        """Testing Bitbucket.authorize with bad credentials"""
        hosting_account = self.create_hosting_account(data={})
        expected_message = (
            'Invalid Bitbucket username or password. Make sure you are using '
            'your Bitbucket username and not e-mail address, and are using an '
            'app password if two-factor authentication is enabled.'
        )

        with self.setup_http_test(status_code=401,
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.is_authorized())

            with self.assertRaisesMessage(AuthorizationError,
                                          expected_message):
                ctx.service.authorize(username='myuser',
                                      password='abc123')

        self.assertNotIn('password', hosting_account.data)
        self.assertFalse(ctx.service.is_authorized())

        ctx.assertHTTPCall(
            0,
            url='https://bitbucket.org/api/2.0/user',
            username='myuser',
            password='abc123')

    def test_authorize_with_403(self):
        """Testing Bitbucket.authorize with HTTP 403 result"""
        hosting_account = self.create_hosting_account(data={})
        expected_message = (
            'Invalid Bitbucket username or password. Make sure you are using '
            'your Bitbucket username and not e-mail address, and are using '
            'an app password if two-factor authentication is enabled.'
        )

        with self.setup_http_test(status_code=403,
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.is_authorized())

            with self.assertRaisesMessage(AuthorizationError,
                                          expected_message):
                ctx.service.authorize(username='myuser',
                                      password='abc123')

        self.assertNotIn('password', hosting_account.data)
        self.assertFalse(ctx.service.is_authorized())

        ctx.assertHTTPCall(
            0,
            url='https://bitbucket.org/api/2.0/user',
            username='myuser',
            password='abc123')

    def test_get_file_with_mercurial_and_base_commit_id(self):
        """Testing Bitbucket.get_file with Mercurial and base commit ID"""
        self._test_get_file(
            tool_name='Mercurial',
            revision='123',
            base_commit_id='456',
            expected_revision='456')

    def test_get_file_with_mercurial_and_revision(self):
        """Testing Bitbucket.get_file with Mercurial and revision"""
        self._test_get_file(
            tool_name='Mercurial',
            revision='123',
            base_commit_id=None,
            expected_revision='123')

    def test_get_file_with_git_and_base_commit_id(self):
        """Testing Bitbucket.get_file with Git and base commit ID"""
        self._test_get_file(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456')

    def test_get_file_with_git_and_revision(self):
        """Testing Bitbucket.get_file with Git and revision"""
        with self.assertRaises(FileNotFoundError):
            self._test_get_file(tool_name='Git',
                                revision='123',
                                base_commit_id=None,
                                expected_revision='123')

    def test_get_file_exists_with_mercurial_and_base_commit_id(self):
        """Testing Bitbucket.get_file_exists with Mercurial and base commit ID
        """
        self._test_get_file_exists(
            tool_name='Mercurial',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=True)

    def test_get_file_exists_with_mercurial_and_revision(self):
        """Testing Bitbucket.get_file_exists with Mercurial and revision"""
        self._test_get_file_exists(
            tool_name='Mercurial',
            revision='123',
            base_commit_id=None,
            expected_revision='123',
            expected_found=True)

    def test_get_file_exists_with_git_and_base_commit_id(self):
        """Testing Bitbucket.get_file_exists with Git and base commit ID"""
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=True)

    def test_get_file_exists_with_git_and_revision(self):
        """Testing Bitbucket.get_file_exists with Git and revision"""
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id=None,
            expected_revision='123',
            expected_found=False,
            expected_http_called=False)

    def test_get_file_exists_with_git_and_404(self):
        """Testing BitBucket.get_file_exists with Git and a 404 error"""
        self._test_get_file_exists(
            tool_name='Git',
            revision='123',
            base_commit_id='456',
            expected_revision='456',
            expected_found=False)

    def test_get_branches(self):
        """Testing Bitbucket.get_branches"""
        branches_api_response_1 = self.dump_json({
            'next': ('https://bitbucket.org/api/2.0/repositories/myuser/'
                     'myrepo/refs/branches?pagelen=100&page=2&'
                     'fields=values.name%2Cvalues.target.hash%2Cnext'),
            'values': [
                {
                    'name': 'branch1',
                    'target': {
                        'hash': '1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                    },
                },
                {
                    'name': 'branch2',
                    'target': {
                        'hash': '44568f7d33647d286691517e6325fea5c7a21d5e',
                    },
                },
            ],
        })

        branches_api_response_2 = self.dump_json({
            'values': [
                {
                    'name': 'branch3',
                    'target': {
                        'hash': 'e5874a857c51a4a13a849a4d1e52d1c44b461ceb',
                    },
                },
                {
                    'name': 'branch4',
                    'target': {
                        'hash': 'd286691517e6325fea5c7a21d5e44568f7d33647',
                    },
                },
            ],
        })

        get_repository_api_response = self.dump_json({
            'mainbranch': {
                'name': 'branch3',
            },
        })

        paths = {
            '/api/2.0/repositories/myuser/myrepo/': {
                'payload': get_repository_api_response,
            },
            ('/api/2.0/repositories/myuser/myrepo/refs/branches'
             '?pagelen=100&fields=values.name%2Cvalues.target.hash%2Cnext'): {
                 'payload': branches_api_response_1,
            },
            ('/api/2.0/repositories/myuser/myrepo/refs/branches'
             '?pagelen=100&page=2&fields=values.name%2Cvalues.target.hash'
             '%2Cnext'): {
                 'payload': branches_api_response_2,
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=3) as ctx:
            repository = self.create_repository(tool_name='Git')
            branches = ctx.service.get_branches(repository)

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo/'
                 '?fields=mainbranch.name'))

        ctx.assertHTTPCall(
            1,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo/'
                 'refs/branches?pagelen=100&fields=values.name'
                 '%2Cvalues.target.hash%2Cnext'))

        ctx.assertHTTPCall(
            2,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo/'
                 'refs/branches?pagelen=100&page=2&fields=values.name'
                 '%2Cvalues.target.hash%2Cnext'))

        self.assertEqual(
            branches,
            [
                Branch(id='branch1',
                       commit='1c44b461cebe5874a857c51a4a13a849a4d1e52d'),
                Branch(id='branch2',
                       commit='44568f7d33647d286691517e6325fea5c7a21d5e'),
                Branch(id='branch3',
                       commit='e5874a857c51a4a13a849a4d1e52d1c44b461ceb',
                       default=True),
                Branch(id='branch4',
                       commit='d286691517e6325fea5c7a21d5e44568f7d33647'),
            ])

    def test_get_commits(self):
        """Testing Bitbucket.get_commits"""
        payload = self.dump_json({
            'values': [
                {
                    'hash': '1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                    'author': {
                        'raw': 'Some User 1 <user1@example.com>',
                    },
                    'date': '2017-01-24T13:11:22+00:00',
                    'message': 'This is commit 1.',
                    'parents': [
                        {
                            'hash': '44568f7d33647d286691517e6325fea5c7a21d5e',
                        },
                    ],
                },
                {
                    'hash': '44568f7d33647d286691517e6325fea5c7a21d5e',
                    'author': {
                        'raw': 'Some User 2 <user2@example.com>',
                    },
                    'date': '2017-01-23T08:09:10+00:00',
                    'message': 'This is commit 2.',
                    'parents': [
                        {
                            'hash': 'e5874a857c51a4a13a849a4d1e52d1c44b461ceb',
                        },
                    ],
                },
            ],
        })

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository(tool_name='Git')
            commits = ctx.service.get_commits(repository)

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo/'
                 'commits?pagelen=20&fields=values.author.raw%2Cvalues.hash'
                 '%2Cvalues.date%2Cvalues.message%2Cvalues.parents.hash'))

        self.assertEqual(
            commits,
            [
                Commit(author_name='Some User 1 <user1@example.com>',
                       date='2017-01-24T13:11:22+00:00',
                       id='1c44b461cebe5874a857c51a4a13a849a4d1e52d',
                       message='This is commit 1.',
                       parent='44568f7d33647d286691517e6325fea5c7a21d5e'),
                Commit(author_name='Some User 2 <user2@example.com>',
                       date='2017-01-23T08:09:10+00:00',
                       id='44568f7d33647d286691517e6325fea5c7a21d5e',
                       message='This is commit 2.',
                       parent='e5874a857c51a4a13a849a4d1e52d1c44b461ceb'),
            ])

        for commit in commits:
            self.assertIsNone(commit.diff)

    def test_get_change(self):
        """Testing BitBucket.get_change"""
        commit_sha = '1c44b461cebe5874a857c51a4a13a849a4d1e52d'
        parent_sha = '44568f7d33647d286691517e6325fea5c7a21d5e'

        paths = {
            '/api/2.0/repositories/myuser/myrepo/commit/%s' % commit_sha: {
                'payload': self.dump_json({
                    'hash': commit_sha,
                    'author': {
                        'raw': 'Some User <user@example.com>',
                    },
                    'date': '2017-01-24T13:11:22+00:00',
                    'message': 'This is a message.',
                    'parents': [{'hash': parent_sha}],
                }),
            },
            '/api/2.0/repositories/myuser/myrepo/diff/%s' % commit_sha: {
                'payload': b'This is a test \xc7.',
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=2) as ctx:
            repository = ctx.create_repository(tool_name='Git')
            commit = ctx.service.get_change(repository, commit_sha)

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo/'
                 'commit/1c44b461cebe5874a857c51a4a13a849a4d1e52d'
                 '?fields=author.raw%2Chash%2Cdate%2Cmessage%2Cparents.hash'))

        ctx.assertHTTPCall(
            1,
            url=('https://bitbucket.org/api/2.0/repositories/myuser/myrepo/'
                 'diff/1c44b461cebe5874a857c51a4a13a849a4d1e52d'))

        self.assertEqual(
            commit,
            Commit(author_name='Some User <user@example.com>',
                   date='2017-01-24T13:11:22+00:00',
                   id=commit_sha,
                   message='This is a message.',
                   parent=parent_sha))
        self.assertEqual(commit.diff, b'This is a test \xc7.\n')

    def _test_get_file(self, tool_name, revision, base_commit_id,
                       expected_revision):
        """Test file fetching.

        Args:
            tool_name (unicode):
                The name of the SCM Tool to test with.

            revision (unicode, optional):
                The revision to check.

            base_commit_id (unicode, optional):
                The base commit to fetch against.

            expected_revision (unicode, optional):
                The revision expected in the payload.
        """
        with self.setup_http_test(payload=b'My data',
                                  expected_http_calls=1) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            result = ctx.service.get_file(repository=repository,
                                          path='path',
                                          revision=revision,
                                          base_commit_id=base_commit_id)

        ctx.assertHTTPCall(
            0,
            url=('https://bitbucket.org/api/1.0/repositories/myuser/myrepo/'
                 'raw/%s/path'
                 % expected_revision))

        self.assertIsInstance(result, bytes)
        self.assertEqual(result, b'My data')

    def _test_get_file_exists(self, tool_name, revision, base_commit_id,
                              expected_revision, expected_found,
                              expected_http_called=True):
        """Test file existence checks.

        Args:
            tool_name (unicode):
                The name of the SCM Tool to test with.

            revision (unicode, optional):
                The revision to check.

            base_commit_id (unicode, optional):
                The base commit to fetch against.

            expected_revision (unicode, optional):
                The revision expected in the payload.

            expected_found (bool, optional):
                Whether a truthy response should be expected.

            expected_http_called (bool, optional):
                Whether an HTTP request is expected to have been made.
        """
        if expected_found:
            payload = b'{}'
            status_code = None
        else:
            payload = b'Not Found'
            status_code = 404

        if expected_http_called:
            expected_calls = 1
        else:
            expected_calls = 0

        with self.setup_http_test(payload=payload,
                                  status_code=status_code,
                                  expected_http_calls=expected_calls) as ctx:
            repository = ctx.create_repository(tool_name=tool_name)
            result = ctx.service.get_file_exists(repository=repository,
                                                 path='path',
                                                 revision=revision,
                                                 base_commit_id=base_commit_id)

        if expected_http_called:
            ctx.assertHTTPCall(
                0,
                url=('https://bitbucket.org/api/1.0/repositories/myuser/'
                     'myrepo/raw/%s/path'
                     % expected_revision))

        self.assertEqual(result, expected_found)


class CloseSubmittedHookTests(BitbucketTestCase):
    """Unit tests for the Bitbucket close-submitted webhook."""

    fixtures = ['test_users', 'test_scmtools']

    def test_close_submitted_hook(self):
        """Testing BitBucket close_submitted hook"""
        self._test_post_commit_hook()

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_with_local_site(self):
        """Testing BitBucket close_submitted hook with a Local Site"""
        self._test_post_commit_hook(
            LocalSite.objects.get(name=self.local_site_name))

    def test_close_submitted_hook_with_invalid_repo(self):
        """Testing BitBucket close_submitted hook with invalid repository"""
        repository = self.create_repository()

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'bitbucket-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'bitbucket',
                'hooks_uuid': repository.get_or_create_hooks_uuid(),
            })

        response = self._post_commit_hook_payload(url, review_request)
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    @add_fixtures(['test_site'])
    def test_close_submitted_hook_with_invalid_site(self):
        """Testing BitBucket close_submitted hook with invalid Local Site"""
        local_site = LocalSite.objects.get(name=self.local_site_name)
        account = self.create_hosting_account(local_site=local_site)
        account.save()

        repository = self.create_repository(hosting_account=account,
                                            local_site=local_site)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'bitbucket-hooks-close-submitted',
            local_site_name='badsite',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'bitbucket',
                'hooks_uuid': repository.get_or_create_hooks_uuid(),
            })

        response = self._post_commit_hook_payload(url, review_request)
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def test_close_submitted_hook_with_invalid_service_id(self):
        """Testing BitBucket close_submitted hook with invalid hosting
        service ID
        """
        # We'll test against GitHub for this test.
        account = self.create_hosting_account()
        account.service_name = 'github'
        account.save()

        repository = self.create_repository(hosting_account=account)

        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'bitbucket-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'bitbucket',
                'hooks_uuid': repository.get_or_create_hooks_uuid(),
            })

        response = self._post_commit_hook_payload(url, review_request)
        self.assertEqual(response.status_code, 404)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)
        self.assertEqual(review_request.changedescs.count(), 0)

    def _test_post_commit_hook(self, local_site=None):
        """Testing posting to a commit hook.

        This will simulate pushing a commit and posting the resulting webhook
        payload from Bitbucket to the handler for the hook.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning the review request.
        """
        account = self.create_hosting_account(local_site=local_site)
        repository = self.create_repository(hosting_account=account,
                                            local_site=local_site)

        review_request = self.create_review_request(repository=repository,
                                                    local_site=local_site,
                                                    publish=True)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.PENDING_REVIEW)

        url = local_site_reverse(
            'bitbucket-hooks-close-submitted',
            local_site=local_site,
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'bitbucket',
                'hooks_uuid': repository.get_or_create_hooks_uuid(),
            })

        self._post_commit_hook_payload(url, review_request)

        review_request = ReviewRequest.objects.get(pk=review_request.pk)
        self.assertTrue(review_request.public)
        self.assertEqual(review_request.status, review_request.SUBMITTED)
        self.assertEqual(review_request.changedescs.count(), 1)

        changedesc = review_request.changedescs.get()
        self.assertEqual(changedesc.text, 'Pushed to master (1c44b46)')

    def _post_commit_hook_payload(self, url, review_request):
        """Post a payload for a hook for testing.

        Args:
            url (unicode):
                The URL to post to.

            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request being represented in the payload.

        Results:
            django.core.handlers.request.wsgi.WSGIRequest:
            The post request.
        """
        return self.client.post(
            url,
            data={
                'payload': self.dump_json({
                    # NOTE: This payload only contains the content we make
                    #       use of in the hook.
                    'commits': [
                        {
                            'raw_node': '1c44b461cebe5874a857c51a4a13a84'
                                        '9a4d1e52d',
                            'branch': 'master',
                            'message': 'This is my fancy commit\n'
                                       '\n'
                                       'Reviewed at http://example.com%s'
                                       % review_request.get_absolute_url(),
                        },
                    ]
                }),
            })



from __future__ import unicode_literals

from django.template.base import add_to_builtins


# This will add the localsite tags as built-in tags, and override the existing
# {% url %} tag in Django.
add_to_builtins(__name__ + '.localsite')



"""Unit tests for reviewboard.reviews.views.BugURLRedirectView."""

from __future__ import unicode_literals

from django.core.urlresolvers import reverse

from reviewboard.testing import TestCase


class BugURLRedirectViewTests(TestCase):
    """Unit tests for reviewboard.reviews.views.BugURLRedirectView."""

    fixtures = ['test_users', 'test_scmtools']

    # Bug #4080
    def test_with_custom_scheme(self):
        """Testing BugURLRedirectView with non-HTTP scheme loads correctly"""
        # Create a repository with a bug tracker that uses a non-standard
        # url scheme.
        repository = self.create_repository(public=True,
                                            bug_tracker='scheme://bugid=%s')
        review_request = self.create_review_request(repository=repository,
                                                    publish=True)
        url = reverse('bug_url', args=(review_request.pk, '1'))
        response = self.client.get(url)

        # Test if we redirected to the correct url with correct bugID.
        self.assertEqual(response['Location'], 'scheme://bugid=1')



"""Unit tests for reviewboard.accounts.forms.pages.ChangePasswordForm."""

from __future__ import unicode_literals

from django.contrib import messages
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.test.client import RequestFactory
from kgb import SpyAgency

from reviewboard.accounts.backends import (AuthBackend,
                                           get_enabled_auth_backends)
from reviewboard.accounts.forms.pages import ChangePasswordForm
from reviewboard.accounts.models import Profile
from reviewboard.testing import TestCase


class SandboxAuthBackend(AuthBackend):
    """Mock authentication backend to test extension sandboxing."""

    backend_id = 'test-id'
    name = 'test'
    supports_change_password = True

    def authenticate(self, username, password):
        """Raise an exception to test sandboxing."""
        raise Exception

    def update_password(self, user, password):
        """Raise an exception to test sandboxing."""
        raise Exception


class ChangePasswordFormTests(SpyAgency, TestCase):
    """Unit tests for reviewboard.accounts.forms.pages.ChangePasswordForm."""

    def setUp(self):
        super(ChangePasswordFormTests, self).setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get('test')
        self.user = User.objects.create_user(username='reviewboard', email='',
                                             password='password')
        self.profile = Profile.objects.get_or_create(user=self.user)
        self.spy_on(get_enabled_auth_backends,
                    call_fake=lambda: [SandboxAuthBackend()])

        # Suppresses MessageFailure Exception at the end of save()
        self.spy_on(messages.add_message,
                    call_fake=lambda *args, **kwargs: None)

    def test_clean_old_password_with_error_in_backend(self):
        """Testing ChangePasswordForm.clean_old_password with error in
        auth_backend.authenticate
        """
        form = ChangePasswordForm(page=None, request=self.request,
                                  user=self.user)
        form.cleaned_data = {
            'old_password': self.user.password,
        }

        self.spy_on(SandboxAuthBackend.authenticate)

        message = (
            'Unexpected error when validating the password. Please contact '
            'the administrator.'
        )

        with self.assertRaisesMessage(ValidationError, message):
            form.clean_old_password()

        self.assertTrue(SandboxAuthBackend.authenticate.called)

    def test_update_password_auth_backend(self):
        """Testing ChangePasswordForm.save with error in
        auth_backend.update_password
        """
        form = ChangePasswordForm(page=None, request=self.request,
                                  user=self.user)
        form.cleaned_data = {
            'old_password': self.user.password,
            'password1': 'password1',
            'password2': 'password1',
        }

        self.spy_on(SandboxAuthBackend.update_password)

        form.save()
        self.assertTrue(SandboxAuthBackend.update_password.called)



"""Admin-specific form widgets."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.forms.widgets import HiddenInput
from django.template.loader import render_to_string
from django.utils import six
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe

from reviewboard.avatars import avatar_services


class RelatedUserWidget(HiddenInput):
    """A form widget to allow people to select one or more User relations.

    It's not unheard of to have a server with thousands or tens of thousands of
    registered users. In this case, the existing Django admin widgets fall down
    hard. The filtered select widgets can actually crash the webserver due to
    trying to pre-populate an enormous ``<select>`` element, and the raw ID
    widget is basically a write-only field.

    This field does much better, offering both the ability to see who's already
    in the list, as well as interactive search and filtering.
    """

    # We inherit from HiddenInput in order for the superclass to render a
    # hidden <input> element, but the siteconfig field template special cases
    # when ``is_hidden`` is True. Setting it to False still gives us the
    # rendering and data handling we want but renders fieldset fields
    # correctly.
    is_hidden = False

    def __init__(self, local_site_name=None, multivalued=True):
        """Initalize the RelatedUserWidget.

        Args:
            local_site_name (unicode, optional):
                The name of the LocalSite where the widget is being rendered.

            multivalued (bool, optional):
                Whether or not the widget should allow selecting multiple
                values.
        """
        super(RelatedUserWidget, self).__init__()

        self.local_site_name = local_site_name
        self.multivalued = multivalued

    def render(self, name, value, attrs=None):
        """Render the widget.

        Args:
            name (unicode):
                The name of the field.

            value (list or None):
                The current value of the field.

            attrs (dict):
                Attributes for the HTML element.

        Returns:
            django.utils.safestring.SafeText:
            The rendered HTML.
        """
        if value:
            if not self.multivalued:
                value = [value]

            value = [v for v in value if v]
            input_value = ','.join(force_text(v) for v in value)
            existing_users = (
                User.objects
                .filter(pk__in=value)
                .order_by('first_name', 'last_name', 'username')
            )
        else:
            input_value = None
            existing_users = []

        final_attrs = self.build_attrs(attrs, name=name)

        input_html = super(RelatedUserWidget, self).render(
            name, input_value, attrs)

        use_avatars = avatar_services.avatars_enabled
        user_data = []

        for user in existing_users:
            data = {
                'fullname': user.get_full_name(),
                'id': user.pk,
                'username': user.username,
            }

            if use_avatars:
                try:
                    data['avatarURL'] = (
                        avatar_services.for_user(user)
                        .get_avatar_urls_uncached(user, 40)
                    )['1x']
                except (AttributeError, KeyError):
                    data['avatarURL'] = None

            user_data.append(data)

        extra_html = render_to_string('admin/related_user_widget.html', {
            'input_id': final_attrs['id'],
            'local_site_name': self.local_site_name,
            'multivalued': self.multivalued,
            'use_avatars': use_avatars,
            'users': user_data,
        })

        return mark_safe(input_html + extra_html)

    def value_from_datadict(self, data, files, name):
        """Unpack the field's value from a datadict.

        Args:
            data (dict):
                The form's data.

            files (dict):
                The form's files.

            name (unicode):
                The name of the field.

        Returns:
            list:
            The list of PKs of User objects.
        """
        value = data.get(name)

        if self.multivalued:
            if isinstance(value, list):
                return value
            elif isinstance(value, six.string_types):
                return [v for v in value.split(',') if v]
            else:
                return None
        elif value:
            return value
        else:
            return None



"""Condition choices and operators for review requests and related objects."""

from __future__ import unicode_literals

from itertools import chain

from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _
from djblets.conditions.choices import (BaseConditionStringChoice,
                                        BaseConditionModelMultipleChoice,
                                        ConditionChoiceMatchListItemsMixin,
                                        ConditionChoices)
from djblets.conditions.operators import (AnyOperator,
                                          BaseConditionOperator,
                                          ConditionOperators,
                                          ContainsAnyOperator,
                                          DoesNotContainAnyOperator,
                                          IsNotOneOfOperator,
                                          IsOneOfOperator,
                                          UnsetOperator)

from reviewboard.reviews.models import Group
from reviewboard.scmtools.conditions import (RepositoriesChoice,
                                             RepositoryTypeChoice)
from reviewboard.site.conditions import LocalSiteModelChoiceMixin


class ReviewRequestConditionChoiceMixin(object):
    """Mixin for condition choices that operate off review requests.

    This will set state needed to match against the choice.
    """

    value_kwarg = 'review_request'


class AnyReviewGroupsPublicOperator(BaseConditionOperator):
    """An operator for matching against any public review groups."""

    operator_id = 'any-public'
    name = _('Are any public')
    value_field = None

    def matches(self, match_value, **kwargs):
        """Return whether any review groups are public.

        Args:
            match_value (list of reviewboard.reviews.models.group.Group):
                The review groups to match.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            bool:
            ``True`` if any review groups are public. ``False`` if all are
            invite-only or the list is empty.
        """
        return match_value and any(
            not group.invite_only
            for group in match_value
        )


class AllReviewGroupsInviteOnlyOperator(BaseConditionOperator):
    """An operator for matching against all invite-only review groups."""

    operator_id = 'all-invite-only'
    name = _('Are all invite-only')
    value_field = None

    def matches(self, match_value, **kwargs):
        """Return whether all review groups in the list are invite-only.

        Args:
            match_value (list of reviewboard.reviews.models.group.Group):
                The review groups to match.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            bool:
            ``True`` if all review group in the list are invite-only.
            ``False`` if any are public, or the list is empty.
        """
        return match_value and all(
            group.invite_only
            for group in match_value
        )


class ReviewGroupsChoice(LocalSiteModelChoiceMixin,
                         BaseConditionModelMultipleChoice):
    """A condition choice for matching review groups.

    This is used to match a :py:class:`~reviewboard.reviews.models.group.Group`
    against a list of groups, against no group (empty list), or against a
    group's public/invite-only state.
    """

    queryset = Group.objects.all()
    choice_id = 'review-groups'
    name = _('Review groups')
    value_kwarg = 'review_groups'

    operators = ConditionOperators([
        AnyOperator.with_overrides(name=_('Any review groups')),
        UnsetOperator.with_overrides(name=_('No review groups')),
        ContainsAnyOperator,
        DoesNotContainAnyOperator,
        AnyReviewGroupsPublicOperator,
        AllReviewGroupsInviteOnlyOperator,
    ])

    def get_match_value(self, review_groups, value_state_cache, **kwargs):
        """Return the review groups used for matching.

        Args:
            review_groups (django.db.models.query.QuerySet):
                The provided queryset for review groups.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            list of reviewboard.reviews.models.group.Group:
            The list of review groups.
        """
        try:
            result = value_state_cache['review_groups']
        except KeyError:
            result = list(review_groups.all())
            value_state_cache['review_groups'] = result

        return result


class ReviewRequestBranchChoice(ReviewRequestConditionChoiceMixin,
                                BaseConditionStringChoice):
    """A condition choice for matching a review request's branch."""

    choice_id = 'branch'
    name = _('Branch')

    def get_match_value(self, review_request, **kwargs):
        """Return the branch text used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            unicode:
            The review request's branch text.
        """
        return review_request.branch


class ReviewRequestDescriptionChoice(ReviewRequestConditionChoiceMixin,
                                     BaseConditionStringChoice):
    """A condition choice for matching a review request's description."""

    choice_id = 'description'
    name = _('Description')

    def get_match_value(self, review_request, **kwargs):
        """Return the description text used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            unicode:
            The review request's description text.
        """
        return review_request.description


class BaseReviewRequestDiffFileChoice(ReviewRequestConditionChoiceMixin,
                                      ConditionChoiceMatchListItemsMixin,
                                      BaseConditionStringChoice):
    """A condition choice for matching affected diff files on a review request.

    This matches against a list of file paths that were added/modified/deleted
    on the latest diffset of a review request.
    """

    def get_match_value(self, review_request, value_state_cache, **kwargs):
        """Return the list of filenames used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            set of unicode:
            The set of filenames.
        """
        try:
            result = value_state_cache['diff_filenames']
        except KeyError:
            try:
                diffset = value_state_cache['latest_diffset']
            except KeyError:
                diffset = review_request.get_latest_diffset()
                value_state_cache['latest_diffset'] = diffset

            result = set(chain.from_iterable(
                diffset.files.values_list('source_file', 'dest_file')))
            value_state_cache['diff_filenames'] = result

        return result


class ReviewRequestAnyDiffFileChoice(BaseReviewRequestDiffFileChoice):
    choice_id = 'any_diffed_file'
    name = _('Any diffed file')
    require_match_all_items = False


class ReviewRequestAllDiffFilesChoice(BaseReviewRequestDiffFileChoice):
    choice_id = 'all_diffed_files'
    name = _('All diffed file')
    require_match_all_items = True


class ReviewRequestOwnerChoice(LocalSiteModelChoiceMixin,
                               ReviewRequestConditionChoiceMixin,
                               BaseConditionModelMultipleChoice):
    """A condition choice for matching a review request's owner."""

    queryset = User.objects.all()
    choice_id = 'owner'
    name = _('Owner')

    operators = ConditionOperators([
        IsOneOfOperator,
        IsNotOneOfOperator,
    ])

    def get_match_value(self, review_request, **kwargs):
        """Return the owner used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            django.contrib.auth.models.User:
            The review request's owner.
        """
        return review_request.owner


class ReviewRequestReviewerChoice(LocalSiteModelChoiceMixin,
                                  ReviewRequestConditionChoiceMixin,
                                  BaseConditionModelMultipleChoice):
    """A condition choice for matching a review request's reviewer."""

    queryset = User.objects.all()
    choice_id = 'reviewer'
    name = _('Reviewer')

    operators = ConditionOperators([
        ContainsAnyOperator,
        DoesNotContainAnyOperator,
    ])

    def get_match_value(self, review_request, **kwargs):
        """Return the reviewers used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            list of django.contrib.auth.models.User:
            List of the review request's reviewers.
        """
        return list(review_request.target_people.all())


class ReviewRequestParticipantChoice(LocalSiteModelChoiceMixin,
                                     ReviewRequestConditionChoiceMixin,
                                     BaseConditionModelMultipleChoice):
    """A condition choice for matching a review request's participant."""

    queryset = User.objects.all()
    choice_id = 'participant'
    name = _('Participant')

    operators = ConditionOperators([
        ContainsAnyOperator,
        DoesNotContainAnyOperator,
    ])

    def get_match_value(self, review_request, **kwargs):
        """Return the participants used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            list of django.contrib.auth.models.User:
            List of the review request's participants.
        """
        return review_request.participants


class ReviewRequestSummaryChoice(ReviewRequestConditionChoiceMixin,
                                 BaseConditionStringChoice):
    """A condition choice for matching a review request's summary."""

    choice_id = 'summary'
    name = _('Summary')

    def get_match_value(self, review_request, **kwargs):
        """Return the summary text used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            unicode:
            The review request's summary text.
        """
        return review_request.summary


class ReviewRequestTestingDoneChoice(BaseConditionStringChoice):
    """A condition choice for matching a review request's Testing Done field.
    """

    choice_id = 'testing-done'
    name = _('Testing Done')

    def get_match_value(self, review_request, **kwargs):
        """Return the testing done text used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            unicode:
            The review request's testing done text.
        """
        return review_request.testing_done


class ReviewRequestRepositoriesChoice(ReviewRequestConditionChoiceMixin,
                                      RepositoriesChoice):
    """A condition choice for matching a review request's repositories."""

    def get_match_value(self, review_request, **kwargs):
        """Return the repository used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            reviewboard.scmtools.models.Repository:
            The review request's repository.
        """
        return review_request.repository


class ReviewRequestRepositoryTypeChoice(ReviewRequestConditionChoiceMixin,
                                        RepositoryTypeChoice):
    """A condition choice for matching a review request's repository types."""

    def get_match_value(self, review_request, **kwargs):
        """Return the repository used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            reviewboard.scmtools.models.Repository:
            The review request's repository.
        """
        return super(ReviewRequestRepositoryTypeChoice, self).get_match_value(
            repository=review_request.repository)


class ReviewRequestReviewGroupsChoice(ReviewRequestConditionChoiceMixin,
                                      ReviewGroupsChoice):
    """A condition choice for matching a review request's review groups."""

    def get_match_value(self, review_request, **kwargs):
        """Return the review groups used for matching.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The provided review request.

            **kwargs (dict):
                Extra keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            The queryset for a review request's target review groups.
        """
        return super(ReviewRequestReviewGroupsChoice, self).get_match_value(
            review_groups=review_request.target_groups,
            **kwargs)


class ReviewRequestConditionChoices(ConditionChoices):
    """A standard set of review request condition choices.

    This provides a handful of condition choices that are useful for
    review requests. They can be used in integrations or any other place
    where conditions are used.
    """

    choice_classes = [
        ReviewRequestBranchChoice,
        ReviewRequestDescriptionChoice,
        ReviewRequestRepositoriesChoice,
        ReviewRequestRepositoryTypeChoice,
        ReviewRequestReviewGroupsChoice,
        ReviewRequestOwnerChoice,
        ReviewRequestReviewerChoice,
        ReviewRequestParticipantChoice,
        ReviewRequestSummaryChoice,
        ReviewRequestTestingDoneChoice,
    ]



"""Unit tests for the Kiln hosting service."""

from __future__ import unicode_literals

from reviewboard.hostingsvcs.errors import RepositoryError
from reviewboard.hostingsvcs.service import HostingServiceClient
from reviewboard.hostingsvcs.testing import HostingServiceTestCase


class KilnTests(HostingServiceTestCase):
    """Unit tests for the Kiln hosting service."""

    service_name = 'kiln'
    fixtures = ['test_scmtools']

    default_account_data = {
        'auth_token': 'my-token',
        'kiln_account_domain': 'mydomain',
    }

    default_repository_extra_data = {
        'kiln_account_domain': 'mydomain',
        'kiln_project_name': 'myproject',
        'kiln_group_name': 'mygroup',
        'kiln_repo_name': 'myrepo',
    }

    def test_service_support(self):
        """Testing Kiln service support capabilities"""
        self.assertTrue(self.service_class.supports_repositories)
        self.assertTrue(self.service_class.needs_authorization)
        self.assertFalse(self.service_class.supports_bug_trackers)
        self.assertFalse(self.service_class.supports_post_commit)
        self.assertFalse(self.service_class.supports_two_factor_auth)

    def test_repo_field_values_git(self):
        """Testing Kiln.get_repository_fields for Git"""
        self.assertEqual(
            self.get_repository_fields(
                'Git',
                fields={
                    'kiln_account_domain': 'mydomain',
                    'kiln_project_name': 'myproject',
                    'kiln_group_name': 'mygroup',
                    'kiln_repo_name': 'myrepo',
                }
            ),
            {
                'path': ('https://mydomain.kilnhg.com/Code/myproject/mygroup/'
                         'myrepo.git'),
                'mirror_path': ('ssh://mydomain@mydomain.kilnhg.com/myproject/'
                                'mygroup/myrepo'),
            })

    def test_repo_field_values_mercurial(self):
        """Testing Kiln.get_repository_fields for Mercurial"""
        self.assertEqual(
            self.get_repository_fields(
                'Mercurial',
                fields={
                    'kiln_account_domain': 'mydomain',
                    'kiln_project_name': 'myproject',
                    'kiln_group_name': 'mygroup',
                    'kiln_repo_name': 'myrepo',
                }
            ),
            {
                'path': ('https://mydomain.kilnhg.com/Code/myproject/mygroup/'
                         'myrepo'),
                'mirror_path': ('ssh://mydomain@mydomain.kilnhg.com/myproject/'
                                'mygroup/myrepo'),
            })

    def test_authorize(self):
        """Testing Kiln.authorize"""
        hosting_account = self.create_hosting_account(data={})

        self.spy_on(HostingServiceClient._make_form_data_boundary,
                    call_fake=lambda: 'BOUNDARY')

        with self.setup_http_test(payload=b'"my-token"',
                                  hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.is_authorized())

            ctx.service.authorize(username='myuser',
                                  password='abc123',
                                  kiln_account_domain='mydomain')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.kilnhg.com/Api/1.0/Auth/Login',
            method='POST',
            username=None,
            password=None,
            body=(
                '--BOUNDARY\r\n'
                'Content-Disposition: form-data; name="sPassword"\r\n\r\n'
                'abc123\r\n'
                '--BOUNDARY\r\n'
                'Content-Disposition: form-data; name="sUser"\r\n\r\n'
                'myuser\r\n'
                '--BOUNDARY--'
            ),
            headers={
                'Content-Length': '152',
                'Content-Type': 'multipart/form-data; boundary=BOUNDARY',
            })

        self.assertIn('auth_token', hosting_account.data)
        self.assertEqual(hosting_account.data['auth_token'], 'my-token')
        self.assertTrue(ctx.service.is_authorized())

    def test_check_repository(self):
        """Testing Kiln.check_repository"""
        payload = self.dump_json([{
            'sSlug': 'myproject',
            'repoGroups': [{
                'sSlug': 'mygroup',
                'repos': [{
                    'sSlug': 'myrepo',
                }]
            }]
        }])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            ctx.service.check_repository(kiln_account_domain='mydomain',
                                         kiln_project_name='myproject',
                                         kiln_group_name='mygroup',
                                         kiln_repo_name='myrepo',
                                         tool_name='Mercurial')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.kilnhg.com/Api/1.0/Project?token=my-token',
            username=None,
            password=None)

    def test_check_repository_with_incorrect_repo_info(self):
        """Testing Kiln.check_repository with incorrect repo info"""
        payload = self.dump_json([{
            'sSlug': 'otherproject',
            'repoGroups': [{
                'sSlug': 'othergroup',
                'repos': [{
                    'sSlug': 'otherrepo',
                }]
            }]
        }])

        expected_message = (
            'The repository with this project, group, and name was not found. '
            'Please verify that the information exactly matches the '
            'configuration on Kiln.'
        )

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(kiln_account_domain='mydomain',
                                             kiln_project_name='myproject',
                                             kiln_group_name='mygroup',
                                             kiln_repo_name='myrepo',
                                             tool_name='Mercurial')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.kilnhg.com/Api/1.0/Project?token=my-token',
            username=None,
            password=None)

    def test_get_file(self):
        """Testing Kiln.get_file"""
        paths = {
            '/Api/1.0/Project': {
                'payload': self.dump_json([{
                    'sSlug': 'myproject',
                    'repoGroups': [{
                        'sSlug': 'mygroup',
                        'repos': [{
                            'sSlug': 'myrepo',
                            'ixRepo': 123,
                        }]
                    }]
                }]),
            },
            '/Api/1.0/Repo/123/Raw/File/2F70617468': {
                'payload': b'My data',
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=2) as ctx:
            repository = ctx.create_repository(tool_name='Mercurial')
            result = ctx.service.get_file(repository=repository,
                                          path='/path',
                                          revision='123')

        self.assertIsInstance(result, bytes)
        self.assertEqual(result, b'My data')

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.kilnhg.com/Api/1.0/Project?token=my-token',
            method='GET',
            username=None,
            password=None)

        ctx.assertHTTPCall(
            1,
            url=('https://mydomain.kilnhg.com/Api/1.0/Repo/123/Raw/File/'
                 '2F70617468?rev=123&token=my-token'),
            username=None,
            password=None)

    def test_get_file_exists(self):
        """Testing Kiln.get_file_exists"""
        paths = {
            '/Api/1.0/Project': {
                'payload': self.dump_json([{
                    'sSlug': 'myproject',
                    'repoGroups': [{
                        'sSlug': 'mygroup',
                        'repos': [{
                            'sSlug': 'myrepo',
                            'ixRepo': 123,
                        }]
                    }]
                }]),
            },
            '/Api/1.0/Repo/123/Raw/File/2F70617468': {
                'payload': b'My data',
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=2) as ctx:
            repository = ctx.create_repository()
            result = ctx.service.get_file_exists(repository=repository,
                                                 path='/path',
                                                 revision='123')

        self.assertTrue(result)

        ctx.assertHTTPCall(
            0,
            url='https://mydomain.kilnhg.com/Api/1.0/Project?token=my-token',
            username=None,
            password=None)

        ctx.assertHTTPCall(
            1,
            url=('https://mydomain.kilnhg.com/Api/1.0/Repo/123/Raw/File/'
                 '2F70617468?rev=123&token=my-token'),
            username=None,
            password=None)



from __future__ import unicode_literals

from reviewboard.reviews.ui.base import ReviewUI
from reviewboard.reviews.models import Screenshot, ScreenshotComment


class LegacyScreenshotReviewUI(ReviewUI):
    name = 'Screenshot'
    model = Screenshot
    comment_model = ScreenshotComment

    js_model_class = 'RB.ScreenshotReviewable'
    js_view_class = 'RB.ImageReviewableView'

    def get_comments(self):
        """Return all existing comments on the screenshot.

        Returns:
            list of reviewboard.reviews.screenshot_comment.ScreenshotComment:
            The list of comments for the page.
        """
        return self.obj.get_comments()

    def get_caption(self, draft=None):
        """Return the caption to show for the screenshot.

        Args:
            draft (reviewboard.reviews.models.review_request_draft.
                   ReviewRequestDraft, optional):
                The active review request draft for the user, if any.

        Returns:
            unicode:
            The caption for the screenshot.
        """
        if draft and self.obj.draft_caption:
            return self.obj.draft_caption

        return self.obj.caption

    def get_comment_link_text(self, comment):
        """Return the text to link to a comment.

        Args:
            comment (reviewboard.reviews.models.screenshot_comment.
                     ScreenshotComment):
                The comment to return text for.

        Returns:
            unicode:
            The text used to link to the comment.
        """
        return self.obj.display_name

    def get_js_model_data(self):
        return {
            'imageURL': self.obj.image.url,
            'screenshotID': self.obj.id,
        }

    def serialize_comments(self, comments):
        result = {}
        serialized_comments = \
            super(LegacyScreenshotReviewUI, self).serialize_comments(comments)

        for serialized_comment in serialized_comments:
            position = '%(x)sx%(y)s+%(w)s+%(h)s' % serialized_comment
            result.setdefault(position, []).append(serialized_comment)

        return result

    def serialize_comment(self, comment):
        data = super(LegacyScreenshotReviewUI, self).serialize_comment(
            comment)

        data.update({
            'x': comment.x,
            'y': comment.y,
            'w': comment.w,
            'h': comment.h,
        })

        return data



"""OAuth initialization.

This module loads the WebAPI scopes when Review Board initializes.
"""

from __future__ import unicode_literals

from djblets.webapi.oauth2_scopes import enable_web_api_scopes
from reviewboard.signals import initializing


default_app_config = 'reviewboard.oauth.apps.OAuthAppConfig'


initializing.connect(enable_web_api_scopes)



from __future__ import unicode_literals

import os

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _
from djblets.util.templatetags.djblets_images import thumbnail

from reviewboard.site.urlresolvers import local_site_reverse


@python_2_unicode_compatible
class Screenshot(models.Model):
    """A screenshot associated with a review request.

    Like diffs, a screenshot can have comments associated with it.
    These comments are of type
    :py:class:`reviewboard.reviews.models.ScreenshotComment`.
    """
    caption = models.CharField(_("caption"), max_length=256, blank=True)
    draft_caption = models.CharField(_("draft caption"),
                                     max_length=256, blank=True)
    image = models.ImageField(_("image"),
                              upload_to=os.path.join('uploaded', 'images',
                                                     '%Y', '%m', '%d'))

    @property
    def filename(self):
        """Returns the filename for display purposes."""
        return os.path.basename(self.image.name)

    def get_comments(self):
        """Returns all the comments made on this screenshot."""
        if not hasattr(self, '_comments'):
            self._comments = list(self.comments.all())

        return self._comments

    def get_thumbnail_url(self):
        """Returns the URL for the thumbnail, creating it if necessary."""
        return thumbnail(self.image)

    def thumb(self):
        """Creates and returns HTML for this screenshot's thumbnail."""
        url = self.get_thumbnail_url()
        return format_html(
            '<img src="{src_1x}" srcset="{src_1x} 1x, {src_2x} 2x"'
            ' alt="{caption}" width="400" />',
            src_1x=url,
            src_2x=thumbnail(self.image, '800x200'),
            caption=self.caption)
    thumb.allow_tags = True

    def __str__(self):
        return "%s (%s)" % (self.caption, self.image)

    def get_review_request(self):
        if hasattr(self, '_review_request'):
            return self._review_request

        try:
            return self.review_request.all()[0]
        except IndexError:
            try:
                return self.inactive_review_request.all()[0]
            except IndexError:
                # Maybe it's on a draft.
                try:
                    draft = self.drafts.get()
                except ObjectDoesNotExist:
                    draft = self.inactive_drafts.get()

                return draft.review_request

    def get_absolute_url(self):
        review_request = self.get_review_request()

        if review_request.local_site:
            local_site_name = review_request.local_site.name
        else:
            local_site_name = None

        return local_site_reverse(
            'screenshot',
            local_site_name=local_site_name,
            kwargs={
                'review_request_id': review_request.display_id,
                'screenshot_id': self.pk,
            })

    def save(self, **kwargs):
        super(Screenshot, self).save()

        try:
            draft = self.drafts.get()
            draft.timestamp = timezone.now()
            draft.save()
        except ObjectDoesNotExist:
            pass

    class Meta:
        app_label = 'reviews'
        db_table = 'reviews_screenshot'
        verbose_name = _('Screenshot')
        verbose_name_plural = _('Screenshots')



"""Database models for integration configuration storage."""

from __future__ import unicode_literals

import logging

from django.db import models
from django.utils.translation import ugettext_lazy as _
from djblets.conditions import ConditionSet
from djblets.integrations.models import BaseIntegrationConfig

from reviewboard.integrations.base import GetIntegrationManagerMixin
from reviewboard.site.models import LocalSite


class IntegrationConfig(GetIntegrationManagerMixin, BaseIntegrationConfig):
    """Stored configuration for a particular integration.

    This contains configuration settings for a given instance of an
    integration, along with state indicating if that integration is to be
    enabled and user-specified identifying information.
    """

    local_site = models.ForeignKey(
        LocalSite,
        related_name='integration_configs',
        blank=True,
        null=True)

    def load_conditions(self, form_cls, conditions_key='conditions'):
        """Load a set of conditions from the configuration.

        This loads and deserializes a set of conditions from the configuration
        stored in the provided key. Those conditions can then be matched by the
        caller.

        If the conditions are not found, this will return ``None``.

        If the conditions cannot be deserialized, this will log some diagnostic
        output and error messages and return ``None``.

        Args:
            form_cls (type):
                The configuration form class that owns the condition field.

                This will generally be ``my_integration.form_cls``,
                but can be another form in more complex integrations.

            conditions_key (unicode, optional):
                The key for the conditions data in the configuration.
                Defaults to "conditions".

        Returns:
            djblets.conditions.conditions.ConditionSet:
            The condition set based on the data, if found and if it could be
            loaded. Otherwise, ``None`` will be returned.
        """
        conditions_data = self.get(conditions_key)

        if not conditions_data:
            return None

        try:
            return ConditionSet.deserialize(
                form_cls.base_fields[conditions_key].choices,
                conditions_data,
                choice_kwargs={
                    'local_site': self.local_site,
                })
        except:
            logging.exception('Unable to load bad condition set data for '
                              'integration configuration ID=%s for key="%s"',
                              self.pk, conditions_key)
            logging.debug('Bad conditions data = %r', conditions_data)

            return None

    def match_conditions(self, form_cls, conditions_key='conditions',
                         **match_kwargs):
        """Filter configurations based on a review request.

        If the configuration contains a ``conditions`` key, and the
        configuration form contains a matching field, this will check
        the conditions for matches against the review request.

        Args:
            form_cls (type):
                The configuration form class that owns the condition field.

                This will generally be ``my_integration.form_cls``,
                but can be another form in more complex integrations.

            conditions_key (unicode, optional):
                The key for the conditions data in the configuration.
                Defaults to "conditions".

            **match_kwargs (dict):
                Keyword arguments to match for the conditions.

        Returns:
            bool:
            ``True`` if the specified conditions in the configuration matches
            the provided keyword arguments. ``False`` if not.
        """
        condition_set = self.load_conditions(form_cls, conditions_key)

        if condition_set:
            try:
                return condition_set.matches(**match_kwargs)
            except Exception as e:
                logging.exception(
                    'Unexpected failure when matching conditions for '
                    'integration configuration ID=%s, config_key=%s, '
                    'match_kwargs=%r: %s',
                    self.pk, conditions_key, match_kwargs, e)

        return False

    class Meta:
        db_table = 'integrations_integrationconfig'
        verbose_name = _('Integration Configuration')
        verbose_name_plural = _('Integration Configurations')



"""E-mail message-related decorators."""

from __future__ import unicode_literals

from functools import wraps

from django.conf import settings
from django.http.response import Http404, HttpResponse


def preview_email(email_builder):
    """Create a view to preview an e-mail message.

    The decorated function should accept the arguments a view would expect,
    i.e. an :py:class:`~django.http.HttpRequest` and the keyword arguments
    generated from the URL. It should return a :py:class:`dict` of keyword
    arguments to be passed to the ``email_builder`` function.

    However, if the decorated function returns a
    :py:class:`~django.http.HttpRequest` that response will be returned
    instead.

    The returned view will **always** return a :http:`404` when
    :django:setting:`DEBUG` is ``False`` (i.e., the view only works when Review
    Board is running as a development server.

    Args:
        email_builder (callable):
            A function that generates an
            :py:class:`~reviewboard.notifications.email.message.EmailMessage`
            from the keyword arguments passed into it.

            This function may also return ``None``, in which case a 404 Not
            Found will be returned.

    Returns:
        callable:
        A view that will generate a preview of the e-mail generated by the
        specified function.
    """
    def decorator(f):
        @wraps(f)
        def decorated(request, message_format, **kwargs):
            if not settings.DEBUG:
                raise Http404

            if message_format not in ('text', 'html'):
                raise Http404

            result = f(request, **kwargs)

            if isinstance(result, HttpResponse):
                return result

            message = email_builder(**result)

            if not message:
                raise Http404

            if message_format == 'text':
                content = message.body
                content_type = 'text/plain'
            else:
                for content, content_type in message.alternatives:
                    if content_type == 'text/html':
                        break
                else:
                    raise Http404

            content_type = '%s; charset=utf-8' % content_type
            return HttpResponse(content, content_type=content_type)

        return decorated

    return decorator



from __future__ import unicode_literals

import hashlib
import hmac
import json
import logging
from collections import defaultdict

from django import forms
from django.conf.urls import url
from django.http import HttpResponse, HttpResponseBadRequest
from django.template.context import RequestContext
from django.template.loader import render_to_string
from django.utils import six
from django.utils.six.moves.urllib.error import HTTPError, URLError
from django.utils.translation import ugettext_lazy as _, ugettext

from reviewboard.admin.server import build_server_url, get_server_url
from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceError)
from reviewboard.hostingsvcs.forms import HostingServiceForm
from reviewboard.hostingsvcs.hook_utils import (close_all_review_requests,
                                                get_repository_for_hook,
                                                get_review_request_id)
from reviewboard.hostingsvcs.service import HostingService
from reviewboard.scmtools.core import Branch, Commit, UNKNOWN
from reviewboard.scmtools.crypto_utils import (decrypt_password,
                                               encrypt_password)
from reviewboard.scmtools.errors import FileNotFoundError, SCMError
from reviewboard.site.urlresolvers import local_site_reverse


logger = logging.getLogger(__name__)


def hook_close_submitted(request, local_site_name=None,
                         repository_id=None,
                         hosting_service_id=None):
    """Close review requests as submitted after a push.

    Args:
        request (django.http.HttpRequest):
            The request from the RB Gateway webhook.

        local_site_name (unicode, optional):
            The local site name, if available.

        repository_id (int, optional):
            The ID of the repository, if available.

        hosting_service_id (unicode, optional):
            The ID of the hosting service.

    Returns:
        django.http.HttpResponse;
        A response for the request.
    """
    hook_event = request.META.get('HTTP_X_RBG_EVENT')

    if hook_event == 'ping':
        return HttpResponse()
    elif hook_event != 'push':
        return HttpResponseBadRequest(
            'Only "ping" and "push" events are supported.')

    repository = get_repository_for_hook(repository_id, hosting_service_id,
                                         local_site_name)

    sig = request.META.get('HTTP_X_RBG_SIGNATURE', '')
    m = hmac.new(bytes(repository.get_or_create_hooks_uuid()),
                 request.body, hashlib.sha1)

    if not hmac.compare_digest(m.hexdigest(), sig):
        return HttpResponseBadRequest('Bad signature.')

    try:
        payload = json.loads(request.body)
    except ValueError as e:
        logging.error('The payload is not in JSON format: %s', e)
        return HttpResponseBadRequest('Invalid payload format.')

    if 'commits' not in payload:
        return HttpResponseBadRequest('Invalid payload; expected "commits".')

    server_url = get_server_url(request=request)
    review_request_ids_to_commits = defaultdict(list)

    for commit in payload['commits']:
        commit_id = commit.get('id')
        commit_message = commit.get('message')
        review_request_id = get_review_request_id(
            commit_message, server_url, commit_id, repository)

        targets = commit['target']

        if 'tags' in targets and targets['tags']:
            target = targets['tags'][0]
        elif 'bookmarks' in targets and targets['bookmarks']:
            target = targets['bookmarks'][0]
        elif 'branch' in targets:
            target = targets['branch']
        else:
            target = ''

        if target:
            target_str = '%s (%s)' % (target, commit_id[:7])
        else:
            target_str = commit_id[:7]

        review_request_ids_to_commits[review_request_id].append(target_str)

    if review_request_ids_to_commits:
        close_all_review_requests(review_request_ids_to_commits,
                                  local_site_name,
                                  repository,
                                  hosting_service_id)

    return HttpResponse()


class ReviewBoardGatewayForm(HostingServiceForm):
    """Hosting service form for Review Board Gateway.

    Provide an additional field on top of the base hosting service form to
    allow specification of the repository name.
    """

    rbgateway_repo_name = forms.CharField(
        label=_('Repository name'),
        max_length=128,
        required=True,
        widget=forms.TextInput(attrs={'size': '60'}),
        help_text=_('The name of the repository. This is the name '
                    'specified in the configuration file for rb-gateway.'))


class ReviewBoardGateway(HostingService):
    """Hosting service support for Review Board Gateway.

    Review Board Gateway is a lightweight self-installed source hosting service
    that currently supports Git repositories.
    """

    name = 'Review Board Gateway'
    form = ReviewBoardGatewayForm
    self_hosted = True
    needs_authorization = True
    supports_repositories = True
    supports_post_commit = True
    supported_scmtools = ['Git', 'Mercurial']

    has_repository_hook_instructions = True

    repository_fields = {
        'Git': {
            'path': '%(hosting_url)s/repos/%(rbgateway_repo_name)s/path',
        },
        'Mercurial': {
            'path': '%(hosting_url)s/repos/%(rbgateway_repo_name)s/path',
        },
    }

    repository_url_patterns = [
        url(r'^hooks/close-submitted/$',
            hook_close_submitted,
            name='rbgateway-hooks-close-submitted'),
    ]

    def check_repository(self, path, *args, **kwargs):
        """Check whether the repository exists."""
        self._api_get(path)

    def authorize(self, username, password, hosting_url, *args, **kwargs):
        """Authorize the Review Board Gateway repository.

        Review Board Gateway uses HTTP Basic Auth, so this will store the
        provided password, encrypted, for use in later API requests.

        Similar to GitLab's API, Review Board Gateway will return a private
        token on session authentication.
        """
        try:
            data, headers = self.client.json_post(
                url=hosting_url + '/session',
                username=username,
                password=password)
        except HTTPError as e:
            if e.code == 401:
                raise AuthorizationError(
                    ugettext('The username or password is incorrect.'))
            elif e.code == 404:
                raise HostingServiceError(
                    ugettext('A Review Board Gateway server was not found at '
                             'the provided URL.'))
            else:
                logger.exception('Failed authorization at %s/session: %s',
                                 hosting_url, e)
            raise

        self.account.data['private_token'] = \
            encrypt_password(data['private_token'])
        self.account.save()

    def is_authorized(self):
        """Determine if the account has supported authorization tokens.

        This will check if we have previously stored a private token for the
        account. It does not validate that the token still works.
        """
        return 'private_token' in self.account.data

    def get_file(self, repository, path, revision, base_commit_id, *args,
                 **kwargs):
        """Get a file from ReviewBoardGateway.

        This will perform an API request to fetch the contents of a file.
        """
        url = self._get_file_url(repository, revision, base_commit_id, path)

        try:
            data, is_new = self._api_get(url)
            return data
        except (HTTPError, URLError) as e:
            if e.code == 404:
                raise FileNotFoundError(path, revision)
            else:
                logger.exception('Failed to get file from %s: %s', url, e)
                raise SCMError(six.text_type(e))

    def get_file_exists(self, repository, path, revision, base_commit_id,
                        *args, **kwargs):
        """Check whether a file exists in ReviewBoardGateway.

        This will perform an API request to fetch the meta_data of a file.
        """
        url = self._get_file_url(repository, revision, base_commit_id, path)

        try:
            self._api_head(url)
            return True
        except (HTTPError, URLError) as e:
            if e.code == 404:
                return False
            else:
                logger.exception('Failed to get file exists from %s: %s',
                                 url, e)
                raise SCMError(six.text_type(e))

    def get_branches(self, repository):
        """Return the branches for the repository.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository to fetch branches for.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The branches returned from the Review Board Gateway API.

        Raises:
            SCMError:
                The branches could not be retrieved from Review Board Gateway.
        """
        url = ('%s/repos/%s/branches' %
               (self.account.hosting_url,
                repository.extra_data['rbgateway_repo_name']))

        tool_name = repository.scmtool_class.name

        if tool_name == 'Git':
            default_branch = 'master'
        elif tool_name == 'Mercurial':
            default_branch = 'default'
        else:
            raise SCMError('Review Board Gateway does not support %s',
                           tool_name)

        try:
            data, headers = self._api_get(url)
            branches = json.loads(data)

            results = []

            for branch in branches:
                results.append(Branch(
                    id=branch['name'],
                    commit=branch['id'],
                    default=(branch['name'] == default_branch)
                ))

            return results
        except Exception as e:
            logger.exception('Failed to get branches from %s: %s', url, e)
            raise SCMError(six.text_type(e))

    def get_commits(self, repository, branch=None, start=None):
        if start is not None:
            url = ('%s/repos/%s/branches/%s/commits?start=%s'
                   % (self.account.hosting_url,
                      repository.extra_data['rbgateway_repo_name'],
                      branch,
                      start))
        else:
            url = ('%s/repos/%s/branches/%s/commits'
                   % (self.account.hosting_url,
                      repository.extra_data['rbgateway_repo_name'],
                      branch))

        try:
            data, headers = self._api_get(url)
            commits = json.loads(data)

            results = []

            for commit in commits:
                results.append(Commit(commit['author'],
                                      commit['id'],
                                      commit['date'],
                                      commit['message'],
                                      commit['parent_id']))

            return results
        except Exception as e:
            logger.exception('Failed to fetch commits from %s: %s', url, e)
            raise SCMError(six.text_type(e))

    def get_change(self, repository, revision):
        url = ('%s/repos/%s/commits/%s'
               % (self.account.hosting_url,
                  repository.extra_data['rbgateway_repo_name'],
                  revision))

        try:
            data, headers = self._api_get(url)
            commit = json.loads(data)

            return Commit(commit['author'],
                          commit['id'],
                          commit['date'],
                          commit['message'],
                          commit['parent_id'],
                          diff=commit['diff'])

        except Exception as e:
            logger.exception('Failed to fetch commit change from %s: %s',
                             url, e)
            raise SCMError(six.text_type(e))

    def get_repository_hook_instructions(self, request, repository):
        """Returns instructions for setting up incoming webhooks.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            repository (reviewboard.scmtools.models.Repository):
                The repository to show instructions for.

        Returns:
            django.http.HttpResponse:
            The hook installation instructions rendered to the contents of an
            HTTP response.
        """
        example_id = 123
        example_url = build_server_url(local_site_reverse(
            'review-request-detail',
            local_site=repository.local_site,
            kwargs={
                'review_request_id': example_id,
            }))

        hook_uuid = repository.get_or_create_hooks_uuid()
        close_url = build_server_url(local_site_reverse(
            'rbgateway-hooks-close-submitted',
            kwargs={
                'repository_id': repository.pk,
                'hosting_service_id': 'rbgateway',
            },
            local_site=repository.local_site))

        return render_to_string(
            'hostingsvcs/rb-gateway/repo_hook_instructions.html',
            RequestContext(request, {
                'example_id': example_id,
                'example_url': example_url,
                'hook_uuid': hook_uuid,
                'close_url': close_url,
                'repo_name': repository.extra_data['rbgateway_repo_name'],
            }))

    def _get_file_url(self, repository, revision, base_commit_id=None,
                      path=None):
        """Get the URL for accessing the contents of a file.

        A revision or a base commit id, path pair is expected to be provided.
        By default, this will return the URL based on the revision, if both
        are provided.
        """
        if revision and revision != UNKNOWN:
            return ('%s/repos/%s/file/%s'
                    % (self.account.hosting_url,
                       repository.extra_data['rbgateway_repo_name'],
                       revision))
        else:
            return ('%s/repos/%s/commits/%s/path/%s'
                    % (self.account.hosting_url,
                       repository.extra_data['rbgateway_repo_name'],
                       base_commit_id,
                       path))

    def _api_get(self, url):
        """Make a GET request to the Review Board Gateway API.

        Delegate to the client's http_get function but first add a
        PRIVATE-TOKEN in the header for authentication.
        """
        try:
            data, headers = self.client.http_get(
                url,
                headers={
                    'PRIVATE-TOKEN': self._get_private_token(),
                })

            return data, headers
        except HTTPError as e:
            if e.code == 401:
                raise AuthorizationError(
                    ugettext('The username or password is incorrect.'))
            elif e.code == 404:
                raise
            else:
                logger.exception('Failed to execute a GET request at %s: %s',
                                 url, e)
                raise

    def _api_head(self, url):
        """Make a HEAD request to the Review Board Gateway API.

        Delegate to the client's http_request function using the method
        HEAD but first add a PRIVATE-TOKEN in the header for authentication.
        """
        try:
            data, headers = self.client.http_request(
                url,
                headers={
                    'PRIVATE-TOKEN': self._get_private_token(),
                },
                method='HEAD')

            return headers
        except HTTPError as e:
            if e.code == 401:
                raise AuthorizationError(
                    ugettext('The username or password is incorrect.'))
            elif e.code == 404:
                raise
            else:
                logger.exception('Failed to execute a HEAD request at %s: %s',
                                 url, e)
                raise

    def _get_private_token(self):
        """Return the private token used for authentication."""
        return decrypt_password(self.account.data['private_token'])



from __future__ import unicode_literals

from djblets.avatars.settings import AvatarSettingsManager


class UserProfileAvatarSettingsManager(AvatarSettingsManager):
    """A mixin that provides avatar service configuration from profiles."""

    AVATAR_SETTINGS_KEY = 'avatars'
    AVATAR_SETTINGS_SERVICE_ID_KEY = 'avatar_service_id'
    AVATAR_SETTINGS_CONFIGURATION_KEY = 'configuration'

    def __init__(self, user):
        super(UserProfileAvatarSettingsManager, self).__init__(user)

        self.profile = user.get_profile()
        self.profile.settings = self.profile.settings or {}
        self.avatar_settings = self.profile.settings.setdefault(
            self.AVATAR_SETTINGS_KEY, {})

    @property
    def avatar_service_id(self):
        """Return service ID for the user's selected avatar service.

        Returns:
            unicode:
            The avatar service ID for the user's selected avatar service, or
            ``None`` if they have not selected one.
        """
        return self.avatar_settings.get(self.AVATAR_SETTINGS_SERVICE_ID_KEY)

    @avatar_service_id.setter
    def avatar_service_id(self, avatar_service_id):
        """Set the avatar service ID for the user.

        Args:
            avatar_service_id (unicode):
                The ID of the :py:class:`avatar service
                <djblets.avatars.services.base.AvatarService>` to set.
        """
        self.avatar_settings[self.AVATAR_SETTINGS_SERVICE_ID_KEY] = \
            avatar_service_id

    @property
    def configuration(self):
        """The user's configuration for all avatar services.

        Returns:
            dict:
            The user's configuration for the avatar service.
        """
        return self.avatar_settings.setdefault(
            self.AVATAR_SETTINGS_CONFIGURATION_KEY, {})

    @configuration.setter
    def configuration(self, settings):
        """Save the user's configuration for all avatar services.

        Args:
            settings (dict):
                The configuration to save.
        """
        self.avatar_settings[self.AVATAR_SETTINGS_CONFIGURATION_KEY] = settings

    def configuration_for(self, avatar_service_id):
        """The user's configuration for the given avatar service.

        Args:
            avatar_service_id (unicode):
                The ID of the :py:class:`avatar service
                <djblets.avatars.service.base.AvatarService>` to get the
                configuration for.
        Returns:
            dict:
            The configuration dictionary for the requested avatar service. It
            will be created if it does not already exist.
        """
        return self.configuration.setdefault(avatar_service_id, {})

    def save(self):
        """Save the user's settings to their profile"""
        self.profile.save(update_fields=('settings',))



"""Middleware for account-related functionality."""

from __future__ import unicode_literals

import pytz
from django.conf import settings
from django.contrib import auth
from django.utils import timezone
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.accounts.backends import X509Backend
from reviewboard.accounts.models import Profile


class TimezoneMiddleware(object):
    """Middleware that activates the user's local timezone."""

    def process_request(self, request):
        """Activate the user's selected timezone for this request."""
        if request.user.is_authenticated():
            try:
                user = request.user.get_profile()
                timezone.activate(pytz.timezone(user.timezone))
            except (Profile.DoesNotExist, pytz.UnknownTimeZoneError):
                pass


class UpdateLastLoginMiddleware(object):
    """Middleware that updates a user's last login time more frequently.

    This will update the user's stored login time if it's been more than 30
    minutes since they last made a request. This helps turn the login time into
    a recent activity time, providing a better sense of how often people are
    actively using Review Board.
    """

    #: The smallest period of time between login time updates.
    UPDATE_PERIOD_SECS = 30 * 60  # 30 minutes

    def process_request(self, request):
        """Process the request and update the login time.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.
        """
        user = request.user

        if user.is_authenticated():
            now = timezone.now()
            delta = now - request.user.last_login

            if delta.total_seconds() >= self.UPDATE_PERIOD_SECS:
                user.last_login = now
                user.save(update_fields=('last_login',))


class X509AuthMiddleware(object):
    """Middleware that authenticates a user using X.509 certificates.

    If Review Board is configured to use the X.509 authentication backend, this
    will automatically authenticate the user using the environment variables
    set by mod_ssl.

    Apache needs to be configured with mod_ssl. For Review Board to be usable
    with X.509 client certificate authentication, the ``SSLVerifyClient``
    configuration directive should be set to ``optional``. This will ensure
    that basic authentication will still work, allowing clients to work with a
    username and password.
    """

    def process_request(self, request):
        """Log in users by their certificate if using X.509 authentication.

        This will only log in a user if the request environment (*not* the
        headers) are populated with a pre-verified username, and the request
        is being handled over HTTPS.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.
        """
        if not request.is_secure():
            return

        siteconfig = SiteConfiguration.objects.get_current()

        if siteconfig.get('auth_backend') != X509Backend.backend_id:
            return

        x509_settings_field = getattr(settings, 'X509_USERNAME_FIELD', None)

        if x509_settings_field:
            x509_field = request.environ.get(x509_settings_field)

            if x509_field:
                user = auth.authenticate(x509_field=x509_field)

                if user:
                    request.user = user
                    auth.login(request, user)



"""Administration UI support for integration configurations."""

from __future__ import unicode_literals

from django.contrib import admin

from reviewboard.integrations.models import IntegrationConfig


class IntegrationConfigAdmin(admin.ModelAdmin):
    list_display = ('integration_id', 'name', 'enabled', 'last_updated')


admin.site.register(IntegrationConfig, IntegrationConfigAdmin)



"""Unit tests for the Gerrit hosting service."""

from __future__ import unicode_literals

import base64

from django.utils.six.moves.urllib.request import (HTTPDigestAuthHandler,
                                                   OpenerDirector)

from reviewboard.hostingsvcs.errors import (AuthorizationError,
                                            HostingServiceAPIError,
                                            HostingServiceError,
                                            RepositoryError)
from reviewboard.hostingsvcs.gerrit import GerritForm
from reviewboard.hostingsvcs.testing import HostingServiceTestCase
from reviewboard.scmtools.core import Branch, Commit
from reviewboard.scmtools.crypto_utils import encrypt_password
from reviewboard.scmtools.errors import FileNotFoundError


class GerritTestCase(HostingServiceTestCase):
    """Base class for Gerrit unit tests."""

    service_name = 'gerrit'

    default_account_data = {
        'authorized': False,
        'gerrit_http_password': encrypt_password('mypass'),
        'gerrit_ssh_port': 1234,
    }

    default_repository_tool_name = 'Git'

    default_repository_extra_data = {
        'gerrit_url': 'http://gerrit.example.com/',
        'gerrit_project_name': 'Project',
    }


class GerritFormTests(GerritTestCase):
    """Unit tests for GerritForm."""

    def test_clean(self):
        """Testing GerritForm.clean"""
        form = GerritForm({
            'gerrit_project_name': 'test-project',
            'gerrit_ssh_port': 12345,
            'gerrit_url': 'http://gerrit.example.com:8080',
        })

        self.assertTrue(form.is_valid())
        self.assertEqual(
            form.cleaned_data,
            {
                'gerrit_domain': 'gerrit.example.com',
                'gerrit_project_name': 'test-project',
                'gerrit_ssh_port': 12345,
                'gerrit_url': 'http://gerrit.example.com:8080/',
            })

    def test_clean_with_errors(self):
        """Testing GerritForm.clean with errors"""
        form = GerritForm({
            'gerrit_url': 'invalid',
        })

        self.assertFalse(form.is_valid())
        self.assertEqual(form.cleaned_data, {})
        self.assertIn('gerrit_project_name', form.errors)
        self.assertIn('gerrit_ssh_port', form.errors)
        self.assertIn('gerrit_url', form.errors)
        self.assertEqual(form.errors['gerrit_project_name'],
                         ['This field is required.'])
        self.assertEqual(form.errors['gerrit_ssh_port'],
                         ['This field is required.'])
        self.assertEqual(form.errors['gerrit_url'],
                         ['Enter a valid URL.'])


class GerritClientTests(GerritTestCase):
    """Unit tests for GerritClient."""

    def setUp(self):
        super(GerritClientTests, self).setUp()

        hosting_account = self.create_hosting_account()
        self.client = hosting_account.service.client

    def test_auth_headers(self):
        """Testing GerritClient.http_request sets auth headers"""
        class DummyResponse(object):
            headers = {}

            def read(self):
                return b''

        def _open(*args, **kwargs):
            _open_args.extend(args)

            return DummyResponse()

        _open_args = []
        self.spy_on(OpenerDirector.open, call_fake=_open)

        self.client.http_request(url='http://gerrit.example.com/',
                                 username='test-user',
                                 password='test-pass')

        opener, request = _open_args
        handler = opener.handlers[0]
        self.assertIsInstance(handler, HTTPDigestAuthHandler)
        self.assertEqual(
            handler.passwd.find_user_password(None,
                                              'http://gerrit.example.com/'),
            ('test-user', 'test-pass'))

        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dGVzdC11c2VyOnRlc3QtcGFzcw==',
            })


class GerritTests(GerritTestCase):
    """Unit tests for the Gerrit hosting service."""

    service_name = 'gerrit'

    default_account_data = {
        'authorized': False,
        'gerrit_http_password': encrypt_password('mypass'),
        'gerrit_ssh_port': 1234,
    }

    default_repository_tool_name = 'Git'

    default_repository_extra_data = {
        'gerrit_url': 'http://gerrit.example.com/',
        'gerrit_project_name': 'Project',
    }

    def test_service_support(self):
        """Testing Gerrit service support capabilities"""
        self.assertFalse(self.service_class.supports_bug_trackers)
        self.assertTrue(self.service_class.supports_repositories)
        self.assertFalse(self.service_class.supports_ssh_key_association)
        self.assertTrue(self.service_class.supports_post_commit)

    def test_authorize(self):
        """Testing Gerrit.authorize"""
        hosting_account = self.create_hosting_account(data={})

        with self.setup_http_test(hosting_account=hosting_account,
                                  expected_http_calls=1) as ctx:
            ctx.service.authorize(
                username='myuser',
                password='mypass',
                hosting_url='',
                credentials={
                    'username': 'myuser',
                    'password': 'mypass',
                },
                local_site_name=None,
                gerrit_url='http://gerrit.example.com')

        self.assertIn('authorized', hosting_account.data)
        self.assertTrue(hosting_account.data['authorized'])

        ctx.assertHTTPCall(0, url='http://gerrit.example.com/a/projects/')

    def test_authorize_with_error(self):
        """Testing Gerrit.authorize handles authentication failure"""
        expected_message = (
            'Unable to authenticate to Gerrit at '
            'http://gerrit.example.com/a/projects/. The username or password '
            'used may be invalid.'
        )

        def _http_request(client, *args, **kwargs):
            raise HostingServiceError('', http_code=401)

        with self.setup_http_test(_http_request, expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(AuthorizationError,
                                          expected_message):
                ctx.service.authorize(
                    username='myuser',
                    password='mypass',
                    hosting_url='',
                    credentials={
                        'username': 'myuser',
                        'password': 'mypass',
                    },
                    local_site_name=None,
                    gerrit_url='http://gerrit.example.com')

        ctx.assertHTTPCall(0, url='http://gerrit.example.com/a/projects/')

        self.assertFalse(ctx.hosting_account.data['authorized'])

    def test_check_repository(self):
        """Testing Gerrit.check_repository"""
        payload = self._make_json_rsp({
            'gerrit-reviewboard': {
                'id': 'gerrit-reviewboard',
                'version': self.service_class.REQUIRED_PLUGIN_VERSION_STR,
            },
        })

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=2) as ctx:
            ctx.service.check_repository(
                gerrit_url='http://gerrit.example.com',
                gerrit_project_name='Project')

        ctx.assertHTTPCall(0,
                           url='http://gerrit.example.com/a/projects/Project')
        ctx.assertHTTPCall(1, url='http://gerrit.example.com/a/plugins/')

    def test_check_repository_with_404(self):
        """Testing Gerrit.check_repository with a non-existent repository"""
        def _http_request(client, *args, **kwargs):
            raise HostingServiceAPIError('', 404)

        expected_message = (
            'The project "Project" does not exist or you do not have access '
            'to it.'
        )

        with self.setup_http_test(_http_request,
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(
                    gerrit_url='http://gerrit.example.com',
                    gerrit_project_name='Project')

        ctx.assertHTTPCall(0,
                           url='http://gerrit.example.com/a/projects/Project')

    def test_check_repository_with_no_plugin(self):
        """Testing Gerrit.check_repository with no plugin"""
        expected_message = (
            'The "gerrit-reviewboard" plugin is not installed on the server. '
            'See https://github.com/reviewboard/gerrit-reviewboard-plugin/ '
            'for installation instructions.'
        )

        with self.setup_http_test(payload=self._make_json_rsp({}),
                                  expected_http_calls=2) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(
                    gerrit_url='http://gerrit.example.com',
                    gerrit_project_name='Project')

        ctx.assertHTTPCall(0,
                           url='http://gerrit.example.com/a/projects/Project')
        ctx.assertHTTPCall(1, url='http://gerrit.example.com/a/plugins/')

    def test_check_repository_with_bad_plugin_version(self):
        """Testing Gerrit.check_repository with an outdated plugin"""
        payload = self._make_json_rsp({
            'gerrit-reviewboard': {
                'id': 'gerrit-reviewboard',
                'version': '0.0.0',
            },
        })

        expected_message = (
            'The "gerrit-reviewboard" plugin on the server is an incompatible '
            'version: found 0.0.0 but version %s or higher is required.'
            % self.service_class.REQUIRED_PLUGIN_VERSION_STR
        )

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=2) as ctx:
            with self.assertRaisesMessage(RepositoryError, expected_message):
                ctx.service.check_repository(
                    gerrit_url='http://gerrit.example.com',
                    gerrit_project_name='Project')

        ctx.assertHTTPCall(0,
                           url='http://gerrit.example.com/a/projects/Project')
        ctx.assertHTTPCall(1, url='http://gerrit.example.com/a/plugins/')

    def test_get_file_exists(self):
        """Testing Gerrit.get_file_exists"""
        blob_id = 'a' * 40

        payload = self._make_json_rsp({
            'blobId': blob_id,
        })

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            self.assertTrue(ctx.service.get_file_exists(
                repository=ctx.create_repository(),
                path='/bogus',
                revision=blob_id))

        ctx.assertHTTPCall(
            0,
            url=('http://gerrit.example.com/a/projects/Project/blobs/%s/'
                 % blob_id))

    def test_get_file_exists_with_404(self):
        """Testing Gerrit.get_file_exists with a non-existant file"""
        def _http_request(client, *args, **kwargs):
            raise HostingServiceAPIError('', http_code=404)

        blob_id = 'a' * 40

        with self.setup_http_test(_http_request, expected_http_calls=1) as ctx:
            self.assertFalse(ctx.service.get_file_exists(
                repository=ctx.create_repository(),
                path='/bogus',
                revision=blob_id))

        ctx.assertHTTPCall(
            0,
            url=('http://gerrit.example.com/a/projects/Project/blobs/%s/'
                 % blob_id))

    def test_get_file(self):
        """Testing Gerrit.get_file"""
        blob_id = 'a' * 40

        with self.setup_http_test(payload=base64.b64encode(b'Hello, world!'),
                                  expected_http_calls=1) as ctx:
            data = ctx.service.get_file(repository=ctx.create_repository(),
                                        path='/bogus',
                                        revision=blob_id)

        self.assertIsInstance(data, bytes)
        self.assertEqual(data, b'Hello, world!')

        ctx.assertHTTPCall(
            0,
            url=('http://gerrit.example.com/a/projects/Project/blobs/%s/'
                 'content/'
                 % blob_id))

    def test_get_file_with_404(self):
        """Testing Gerrit.get_file with a non-existent blob ID"""
        def _http_request(client, *args, **kwargs):
            raise HostingServiceAPIError('', http_code=404)

        blob_id = 'a' * 40

        with self.setup_http_test(_http_request, expected_http_calls=1) as ctx:
            with self.assertRaises(FileNotFoundError):
                ctx.service.get_file(repository=ctx.create_repository(),
                                     path='/bogus',
                                     revision=blob_id)

        ctx.assertHTTPCall(
            0,
            url=('http://gerrit.example.com/a/projects/Project/blobs/%s/'
                 'content/'
                 % blob_id))

    def test_get_file_with_undecodable_response(self):
        """Testing Gerrit.get_file with an undecodable response"""
        blob_id = 'a' * 40
        expected_message = (
            'An error occurred while retrieving "/foo" at revision '
            '"aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" from Gerrit: the '
            'response could not be decoded: Incorrect padding'
        )

        with self.setup_http_test(payload=b'?Invalid base64',
                                  expected_http_calls=1) as ctx:
            with self.assertRaisesMessage(HostingServiceAPIError,
                                          expected_message):
                ctx.service.get_file(repository=ctx.create_repository(),
                                     path='/foo',
                                     revision=blob_id)

        ctx.assertHTTPCall(
            0,
            url=('http://gerrit.example.com/a/projects/Project/blobs/%s/'
                 'content/'
                 % blob_id))

    def test_get_branches(self):
        """Testing Gerrit.get_branches"""
        payload = self._make_json_rsp([
            {
                'ref': 'HEAD',
                'revision': 'master'
            },
            {
                'ref': 'refs/meta/config',
                'revision': '7a59a483aeefc8c7d4082f1081c42db817176071'
            },
            {
                'ref': 'refs/heads/master',
                'revision': '6854734ef5fc8b2b9d291bf42aa59c344abf5a73'
            },
            {
                'ref': 'refs/heads/release-2.0.x',
                'revision': '7f68a001f8e5b77e7355c11385bfbcd2a6d3c077'
            },
            {
                'ref': 'refs/heads/release-2.5.x',
                'revision': 'fc8a7ecf288d835ecd9ded086ffaee9412d1da9c'
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            branches = ctx.service.get_branches(ctx.create_repository())

        ctx.assertHTTPCall(
            0,
            url='http://gerrit.example.com/a/projects/Project/branches/')

        self.assertEqual(
            branches,
            [
                Branch(commit='6854734ef5fc8b2b9d291bf42aa59c344abf5a73',
                       id='master',
                       default=True),
                Branch(commit='7f68a001f8e5b77e7355c11385bfbcd2a6d3c077',
                       id='release-2.0.x'),
                Branch(commit='fc8a7ecf288d835ecd9ded086ffaee9412d1da9c',
                       id='release-2.5.x'),
            ]
        )

    def test_get_commits(self):
        """Testing Gerrit.get_commits"""
        payload = self._make_json_rsp([
            {
                'message': 'Backport a fix for screenshot commenting.\n',
                'revision': '77c174669b7018936f16b98547445624c6738e1e',
                'author': 'David Trowbridge',
                'parents': [
                    'ecfbf578d31f550a135580cee26fa20fbaea36d9',
                ],
                'time': '2016-09-05T23:28:30-07:00',
            },
            {
                'message': 'Fix some issues with screenshot commenting.'
                           '\n',
                'revision': '8a39b87f0124f27225234014a87914e434b223a9',
                'author': 'David Trowbridge',
                'parents': [
                    '3fb32c83993cd8c07fbbb605cf0cc523010da7c8',
                ],
                'time': '2016-09-05T22:58:29-07:00',
            },
            {
                'message': 'Fix draggability of the comment dialog.\n',
                'revision': '3fb32c83993cd8c07fbbb605cf0cc523010da7c8',
                'author': 'David Trowbridge',
                'parents': [
                    '7619f51371b55bfcdf4cb3fccf5d3c76bf5002c0',
                ],
                'time': '2016-09-05T22:47:55-07:00',
            },
        ])

        with self.setup_http_test(payload=payload,
                                  expected_http_calls=1) as ctx:
            commits = ctx.service.get_commits(ctx.create_repository())

        ctx.assertHTTPCall(
            0,
            url='http://gerrit.example.com/a/projects/Project/all-commits/')

        self.assertEqual(
            commits,
            [
                Commit(author_name='David Trowbridge',
                       id='77c174669b7018936f16b98547445624c6738e1e',
                       date='2016-09-05T23:28:30-07:00',
                       message='Backport a fix for screenshot commenting.\n',
                       parent='ecfbf578d31f550a135580cee26fa20fbaea36d9'),
                Commit(author_name='David Trowbridge',
                       id='8a39b87f0124f27225234014a87914e434b223a9',
                       date='2016-09-05T22:58:29-07:00',
                       message='Fix some issues with screenshot commenting.\n',
                       parent='3fb32c83993cd8c07fbbb605cf0cc523010da7c8'),
                Commit(author_name='David Trowbridge',
                       id='3fb32c83993cd8c07fbbb605cf0cc523010da7c8',
                       date='2016-09-05T22:47:55-07:00',
                       message='Fix draggability of the comment dialog.\n',
                       parent='7619f51371b55bfcdf4cb3fccf5d3c76bf5002c0'),
            ])

        for commit in commits:
            self.assertIsNone(commit.diff)

    def test_get_change(self):
        """Testing Gerrit.get_change"""
        revision = '77c174669b7018936f16b98547445624c6738e1e'
        paths = {
            '/a/projects/Project/commits/%s/diff/' % revision: {
                'payload': b'fake diff',
            },
            '/a/projects/Project/all-commits/': {
                'payload': self._make_json_rsp([{
                    'message': 'Backport a fix for screenshot '
                               'commenting.\n',
                    'revision': '77c174669b7018936f16b98547445624c6738e1e',
                    'author': 'David Trowbridge',
                    'parents': [
                        'ecfbf578d31f550a135580cee26fa20fbaea36d9',
                    ],
                    'time': '2016-09-05T23:28:30-07:00',
                }]),
            },
        }

        with self.setup_http_test(self.make_handler_for_paths(paths),
                                  expected_http_calls=2) as ctx:
            commit = ctx.service.get_change(
                repository=ctx.create_repository(),
                revision='77c174669b7018936f16b98547445624c6738e1e')

        self.assertEqual(
            commit,
            Commit(author_name='David Trowbridge',
                   id='77c174669b7018936f16b98547445624c6738e1e',
                   date='2016-09-05T23:28:30-07:00',
                   message='Backport a fix for screenshot commenting.\n',
                   parent='ecfbf578d31f550a135580cee26fa20fbaea36d9'))
        self.assertEqual(commit.diff, b'fake diff')

        ctx.assertHTTPCall(
            0,
            url=('http://gerrit.example.com/a/projects/Project/all-commits/'
                 '?start=77c174669b7018936f16b98547445624c6738e1e&limit=1'))

        ctx.assertHTTPCall(
            1,
            url=('http://gerrit.example.com/a/projects/Project/commits/'
                 '77c174669b7018936f16b98547445624c6738e1e/diff/'))

    def _make_json_rsp(self, data):
        """Return a Gerrit JSON response payload for the given data.

        Args:
            data (object):
                The data to serialize.

        Returns:
            bytes:
            The serialized payload data.
        """
        return b")]}'\n%s" % self.dump_json(data)
